import serial
from psychopy import parallel
import numpy as np


class NeuroScanPort:
    """Send tag communication.
    -author: Lichao Xu
    -Created on: 2020-07-30
    -update log:
        None
    Parameters
    ----------
        port_addr: ndarray,
            The port address.
        use_serial: bool,
            Send tags using serial port.
        baudrate: int,
            The serial port baud rate.
    """

    def __init__(self, port_addr, use_serial=False, baudrate=115200):
        self.use_serial = use_serial
        if use_serial:
            self.port = serial.Serial(port=port_addr, baudrate=baudrate)
        else:
            self.port = parallel.ParallelPort(address=port_addr)

    def setData(self, label):
        if self.use_serial:
            self.port.write(int(label))
        else:
            self.port.setData(int(label))


def _check_array_like(value, length=None):
    """Check array dimensions.
    -author: Lichao Xu
    -Created on: 2020-07-30
    -update log:
        None
    Parameters
    ----------
        value: ndarray,
            The array to check.
        length: int,
            The array dimension.
    """

    flag = isinstance(value, (list, tuple, np.ndarray))
    return flag and (len(value) == length if length is not None else True)


def _clean_dict(old_dict, includes=[]):
    """Clear dictionary.
    -author: Lichao Xu
    -Created on: 2020-07-30
    -update log:
        None
    Parameters
    ----------
        old_dict: dict,
            The dict to clear.
        includes: list,
            Key-value indexes that need to be preserved.
    """

    names = list(old_dict.keys())
    for name in names:
        if name not in includes:
            old_dict[name] = None
            del old_dict[name]
    return old_dict



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/01/07
# License: MIT License
"""
Nakanishi SSVEP dataset.
"""
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

import numpy as np
from mne import create_info
from mne.io import RawArray, Raw
from mne.channels import make_standard_montage
from .base import BaseDataset
from ..utils.download import mne_data_path
from ..utils.channels import upper_ch_names
from ..utils.io import loadmat

Nakanishi2015_URL = "https://github.com/mnakanishi/12JFPM_SSVEP/raw/master/data/"


class Nakanishi2015(BaseDataset):
    """SSVEP Nakanishi 2015 dataset

    This dataset contains 12-class joint frequency-phase modulated steady-state
    visual evoked potentials (SSVEPs) acquired from 10 subjects used to
    estimate an online performance of brain-computer interface (BCI) in the
    reference study [1]_.

    references
    ----------
    .. [1] Masaki Nakanishi, Yijun Wang, Yu-Te Wang and Tzyy-Ping Jung,
    "A Comparison Study of Canonical Correlation Analysis Based Methods for
    Detecting Steady-State Visual Evoked Potentials," PLoS One, vol.10, no.10,
    e140703, 2015.
    http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0140703
    """

    _CHANNELS = ["PO7", "PO3", "POZ", "PO4", "PO8", "O1", "OZ", "O2"]

    _FREQS = [
        9.25,
        11.25,
        13.25,
        9.75,
        11.75,
        13.75,
        10.25,
        12.25,
        14.25,
        10.75,
        12.75,
        14.75,
    ]
    _PHASES = [0, 0, 0, 0.5, 0.5, 0.5, 1, 1, 1, 1.5, 1.5, 1.5]

    _EVENTS = {str(freq): (i + 1, (0, 4)) for i, freq in enumerate(_FREQS)}

    def __init__(self):
        super().__init__(
            dataset_code="nakanishi2015",
            subjects=list(range(1, 11)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=256,
            paradigm="ssvep",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        url = "{:s}s{:d}.mat".format(Nakanishi2015_URL, subject)
        file_dest = mne_data_path(
            url,
            self.dataset_code,
            path=path,
            proxies=proxies,
            force_update=force_update,
            update_path=update_path,
        )

        dests = [[file_dest]]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        dests = self.data_path(subject)
        raw_mat = loadmat(dests[0][0])
        n_samples, n_channels, n_trials = 1114, 8, 15
        n_classes = 12

        data = np.transpose(raw_mat["eeg"], axes=(0, 3, 1, 2))
        data = np.reshape(data, newshape=(-1, n_channels, n_samples))
        data = data - data.mean(axis=2, keepdims=True)
        raw_events = np.zeros((data.shape[0], 1, n_samples))
        raw_events[:, 0, 38] = np.array(
            [n_trials * [i + 1] for i in range(n_classes)]
        ).flatten()
        data = np.concatenate([1e-6 * data, raw_events], axis=1)

        buff = (data.shape[0], n_channels + 1, 50)
        data = np.concatenate([np.zeros(buff), data, np.zeros(buff)], axis=2)
        ch_names = self._CHANNELS + ["stim"]
        ch_types = ["eeg"] * len(self._CHANNELS) + ["stim"]

        info = create_info(ch_names=ch_names, ch_types=ch_types, sfreq=self.srate)
        raw = RawArray(data=np.concatenate(list(data), axis=1), info=info)
        raw = upper_ch_names(raw)
        raw.set_montage(montage)

        sess = {"session_0": {"run_0": raw}}
        return sess

    def get_freq(self, event: str):
        return self._FREQS[self._EVENTS[event][0] - 1]

    def get_phase(self, event: str):
        return self._PHASES[self._EVENTS[event][0] - 1]



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/1/07
# License: MIT License
"""
Common Spatial Patterns and his happy little buddies!

"""
from typing import Optional, List, Tuple
from functools import partial

import numpy as np
from numpy import ndarray
from scipy.linalg import eigh
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import GridSearchCV, StratifiedKFold, ShuffleSplit
from sklearn.feature_selection import SelectKBest, mutual_info_classif
from sklearn.svm import SVC
from sklearn.linear_model import Ridge
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier
from sklearn.pipeline import make_pipeline

from .base import robust_pattern, FilterBank
from ..utils.covariance import nearestPD, covariances


def csp_kernel(X: ndarray, y: ndarray) -> Tuple[ndarray, ndarray, ndarray]:
    """The kernel in CSP algorithm based on paper [1]_.

    Parameters
    ----------
    X: ndarray
        eeg data, shape (n_trials, n_channels, n_samples).
    y: ndarray
        labels of X, shape (n_trials,).

    Returns
    -------
    W: ndarray
        Spatial filters, shape (n_channels, n_filters).
    D: ndarray
        Eigenvalues of spatial filters, shape (n_filters,).
    A: ndarray
        Spatial patterns, shape (n_channels, n_patterns).

    References
    ----------
    .. [1] Ramoser H, Muller-Gerking J, Pfurtscheller G. Optimal spatial filtering of single trial EEG during imagined hand
           movement[J]. IEEE transactions on rehabilitation engineering, 2000, 8(4): 441-446.
    """
    X, y = np.copy(X), np.copy(y)
    labels = np.unique(y)
    X = X - np.mean(X, axis=-1, keepdims=True)
    if len(labels) != 2:
        raise ValueError("the current kernel is for 2-class problem.")

    C1 = covariances(X[y == labels[0]])
    C2 = covariances(X[y == labels[1]])

    # # trace normalization
    # # this operation equals to trial normalization
    # C1 = C1 / np.trace(C1, axis1=-1, axis2=-2)[:, np.newaxis, np.newaxis]
    # C2 = C2 / np.trace(C2, axis1=-1, axis2=-2)[:, np.newaxis, np.newaxis]

    C1 = np.mean(C1, axis=0)
    C2 = np.mean(C2, axis=0)
    Cc = C1 + C2
    # check positive-definiteness
    Cc = nearestPD(Cc)
    # generalized eigenvalue problem
    D, W = eigh(C1, Cc)
    ix = np.argsort(D)[::-1]
    W = W[:, ix]
    D = D[ix]

    A = robust_pattern(W, C1, W.T @ C1 @ W)

    return W, D, A


def csp_feature(W: ndarray, X: ndarray, n_components: int = 2) -> ndarray:
    """Return CSP features in paper [1]_.

    Parameters
    ----------
    W : ndarray
        spatial filters from csp_kernel, shape (n_channels, n_filters)
    X : ndarray
        eeg data, shape (n_trials, n_channels, n_samples)
    n_components : int, optional
        the first k components to use, usually even number, by default 2

    Returns
    -------
    ndarray
        features of shape (n_trials, n_features)

    Raises
    ------
    ValueError
        n_components should less than the number of channels

    References
    ----------
    .. [1] Ramoser H, Muller-Gerking J, Pfurtscheller G. Optimal spatial filtering of single trial EEG during imagined hand
           movement[J]. IEEE transactions on rehabilitation engineering, 2000, 8(4): 441-446.
    """
    W, X = np.copy(W), np.copy(X)
    max_components = W.shape[1]
    if n_components > max_components:
        raise ValueError("n_components should less than the number of channels")

    eps = np.finfo(X.dtype).eps
    X = X - np.mean(X, axis=-1, keepdims=True)
    # normalized variance
    features = np.mean(np.square(np.matmul(W[:, :n_components].T, X)), axis=-1)
    features = features / (np.sum(features, axis=-1, keepdims=True) + eps)
    # log-transformation
    features = np.log(np.clip(features, eps, None))
    return features


def _rjd(X, eps=1e-9, n_iter_max=1000):
    """Approximate joint diagonalization based on jacobi angle.

    Parameters
    ----------
    X : ndarray
        A set of covariance matrices to diagonalize, shape (n_trials, n_channels, n_channels).
    eps : float, optional
        Tolerance for stopping criterion (default 1e-8).
    n_iter_max : int, optional
        The maximum number of iteration to reach convergence (default 1000).

    Returns
    -------
    V : ndarray
        The diagonalizer, shape (n_channels, n_filters), usually n_filters == n_channels.
    D : ndarray
        The set of quasi diagonal matrices, shape (n_trials, n_channels, n_channels).

    Notes
    -----
    This is a direct implementation of the Cardoso AJD algorithm [1]_ used in
    JADE. The code is a translation of the matlab code provided in the author
    website.

    References
    ----------
    .. [1] Cardoso, Jean-Francois, and Antoine Souloumiac. Jacobi angles for simultaneous diagonalization. SIAM journal on
           matrix analysis and applications 17.1 (1996): 161-164.

    """

    # reshape input matrix
    A = np.concatenate(X, 0).T

    # init variables
    m, nm = A.shape
    V = np.eye(m)
    encore = True
    k = 0

    while encore:
        encore = False
        k += 1
        if k > n_iter_max:
            break
        for p in range(m - 1):
            for q in range(p + 1, m):

                Ip = np.arange(p, nm, m)
                Iq = np.arange(q, nm, m)

                # computation of Givens angle
                g = np.array([A[p, Ip] - A[q, Iq], A[p, Iq] + A[q, Ip]])
                gg = np.dot(g, g.T)
                ton = gg[0, 0] - gg[1, 1]
                toff = gg[0, 1] + gg[1, 0]
                theta = 0.5 * np.arctan2(toff, ton + np.sqrt(ton * ton + toff * toff))
                c = np.cos(theta)
                s = np.sin(theta)
                encore = encore | (np.abs(s) > eps)
                if np.abs(s) > eps:
                    tmp = A[:, Ip].copy()
                    A[:, Ip] = c * A[:, Ip] + s * A[:, Iq]
                    A[:, Iq] = c * A[:, Iq] - s * tmp

                    tmp = A[p, :].copy()
                    A[p, :] = c * A[p, :] + s * A[q, :]
                    A[q, :] = c * A[q, :] - s * tmp

                    tmp = V[:, p].copy()
                    V[:, p] = c * V[:, p] + s * V[:, q]
                    V[:, q] = c * V[:, q] - s * tmp

    D = np.reshape(A, (m, int(nm / m), m)).transpose(1, 0, 2)
    return V, D


def _ajd_pham(X, eps=1e-9, n_iter_max=1000):
    """Approximate joint diagonalization based on pham's algorithm.

    Parameters
    ----------
    X : ndarray
        A set of covariance matrices to diagonalize, shape (n_trials, n_channels, n_channels).
    eps : float, optional
        Tolerance for stoping criterion (default 1e-6).
    n_iter_max : int, optional
        The maximum number of iteration to reach convergence (default 1000).

    Returns
    -------
    V : ndarray
        The diagonalizer, shape (n_channels, n_filters), usually n_filters == n_channels.
    D : ndarray
        The set of quasi diagonal matrices, shape (n_trials, n_channels, n_channels).

    Notes
    -----
    This is a direct implementation of the PHAM's AJD algorithm [1]_.

    References
    ----------
    .. [1] Pham, Dinh Tuan. "Joint approximate diagonalization of positive definite Hermitian matrices." SIAM Journal on
           Matrix Analysis and Applications 22, no. 4 (2001): 1136-1152.

    """
    # Adapted from http://github.com/alexandrebarachant/pyRiemann
    n_epochs = X.shape[0]

    # Reshape input matrix
    A = np.concatenate(X, axis=0).T

    # Init variables
    n_times, n_m = A.shape
    V = np.eye(n_times)
    epsilon = n_times * (n_times - 1) * eps

    for it in range(n_iter_max):
        decr = 0
        for ii in range(1, n_times):
            for jj in range(ii):
                Ii = np.arange(ii, n_m, n_times)
                Ij = np.arange(jj, n_m, n_times)

                c1 = A[ii, Ii]
                c2 = A[jj, Ij]

                g12 = np.mean(A[ii, Ij] / c1)
                g21 = np.mean(A[ii, Ij] / c2)

                omega21 = np.mean(c1 / c2)
                omega12 = np.mean(c2 / c1)
                omega = np.sqrt(omega12 * omega21)

                tmp = np.sqrt(omega21 / omega12)
                tmp1 = (tmp * g12 + g21) / (omega + 1)
                tmp2 = (tmp * g12 - g21) / max(omega - 1, 1e-9)

                h12 = tmp1 + tmp2
                h21 = np.conj((tmp1 - tmp2) / tmp)

                decr += n_epochs * (g12 * np.conj(h12) + g21 * h21) / 2.0

                tmp = 1 + 1.0j * 0.5 * np.imag(h12 * h21)
                tmp = np.real(tmp + np.sqrt(tmp**2 - h12 * h21))
                tau = np.array([[1, -h12 / tmp], [-h21 / tmp, 1]])

                A[[ii, jj], :] = np.dot(tau, A[[ii, jj], :])
                tmp = np.c_[A[:, Ii], A[:, Ij]]
                tmp = np.reshape(tmp, (n_times * n_epochs, 2), order="F")
                tmp = np.dot(tmp, tau.T)

                tmp = np.reshape(tmp, (n_times, n_epochs * 2), order="F")
                A[:, Ii] = tmp[:, :n_epochs]
                A[:, Ij] = tmp[:, n_epochs:]
                V[[ii, jj], :] = np.dot(tau, V[[ii, jj], :])
        if decr < epsilon:
            break
    D = np.reshape(A, (n_times, -1, n_times)).transpose(1, 0, 2)
    return V.T, D


def _uwedge(X, init=None, eps=1e-9, n_iter_max=1000):
    """Approximate joint diagonalization algorithm UWEDGE.

    Parameters
    ----------
    X : ndarray
        A set of covariance matrices to diagonalize, shape (n_trials, n_channels, n_channels).
    init : None | ndarray, optional
        Initialization for the diagonalizer, shape (n_channels, n_channels).
    eps : float, optional
        Tolerance for stoping criterion (default 1e-7).
    n_iter_max : int
        The maximum number of iteration to reach convergence (default 1000).

    Returns
    -------
    W_est : ndarray
        The diagonalizer, shape (n_filters, n_channels), usually n_filters == n_channels.
    D : ndarray
        The set of quasi diagonal matrices, shape (n_trials, n_channels, n_channels).

    Notes
    -----
    Uniformly Weighted Exhaustive Diagonalization using Gauss iteration
    (U-WEDGE). Implementation of the AJD algorithm by Tichavsky and Yeredor [1]_ [2]_.
    This is a translation from the matlab code provided by the authors.

    References
    ----------
    .. [1] P. Tichavsky, A. Yeredor and J. Nielsen, "A Fast Approximate Joint Diagonalization Algorithm Using a Criterion with
           a Block Diagonal Weight Matrix", ICASSP 2008, Las Vegas.
    .. [2] P. Tichavsky and A. Yeredor, "Fast Approximate Joint Diagonalization Incorporating Weight Matrices" IEEE
           Transactions of Signal Processing, 2009.

    """
    L, d, _ = X.shape

    # reshape input matrix
    M = np.concatenate(X, 0).T

    # init variables
    d, Md = M.shape
    iteration = 0
    improve = 10

    if init is None:
        E, H = np.linalg.eig(M[:, 0:d])
        W_est = np.dot(np.diag(1.0 / np.sqrt(np.abs(E))), H.T)
    else:
        W_est = init

    Ms = np.array(M)
    Rs = np.zeros((d, L))

    for k in range(L):
        ini = k * d
        Il = np.arange(ini, ini + d)
        M[:, Il] = 0.5 * (M[:, Il] + M[:, Il].T)
        Ms[:, Il] = np.dot(np.dot(W_est, M[:, Il]), W_est.T)
        Rs[:, k] = np.diag(Ms[:, Il])

    crit = np.sum(Ms**2) - np.sum(Rs**2)
    while (improve > eps) & (iteration < n_iter_max):
        B = np.dot(Rs, Rs.T)
        C1 = np.zeros((d, d))
        for i in range(d):
            C1[:, i] = np.sum(Ms[:, i:Md:d] * Rs, axis=1)

        D0 = B * B.T - np.outer(np.diag(B), np.diag(B))
        A0 = (C1 * B - np.dot(np.diag(np.diag(B)), C1.T)) / (D0 + np.eye(d))
        A0 += np.eye(d)
        W_est = np.linalg.solve(A0, W_est)

        Raux = np.dot(np.dot(W_est, M[:, 0:d]), W_est.T)
        aux = 1.0 / np.sqrt(np.abs(np.diag(Raux)))
        W_est = np.dot(np.diag(aux), W_est)

        for k in range(L):
            ini = k * d
            Il = np.arange(ini, ini + d)
            Ms[:, Il] = np.dot(np.dot(W_est, M[:, Il]), W_est.T)
            Rs[:, k] = np.diag(Ms[:, Il])

        crit_new = np.sum(Ms**2) - np.sum(Rs**2)
        improve = np.abs(crit_new - crit)
        crit = crit_new
        iteration += 1

    D = np.reshape(Ms, (d, L, d)).transpose(1, 0, 2)
    return W_est.T, D


ajd_methods = {"rjd": _rjd, "ajd_pham": _ajd_pham, "uwedge": _uwedge}


def _check_ajd_method(method):
    """Check if a given method is valid.

    Parameters
    ----------
    method : callable object or str
        Could be the name of ajd_method or a callable method itself.

    Returns
    -------
    method: callable object
        A callable ajd method.
    """
    if callable(method):
        pass
    elif method in ajd_methods.keys():
        method = ajd_methods[method]
    else:
        raise ValueError(
            """%s is not an valid method ! Valid methods are : %s or a
             callable function"""
            % (method, (" , ").join(ajd_methods.keys()))
        )
    return method


def ajd(X: ndarray, method: str = "uwedge") -> Tuple[ndarray, ndarray]:
    """Wrapper of AJD methods.

    Parameters
    ----------
    X : ndarray
        Input covariance matrices, shape (n_trials, n_channels, n_channels)
    method : str, optional
        AJD method (default uwedge).

    Returns
    -------
    V : ndarray
        The diagonalizer, shape (n_channels, n_filters), usually n_filters == n_channels.
    D : ndarray
        The mean of quasi diagonal matrices, shape (n_channels,).
    """
    method_func = _check_ajd_method(method)
    V, D = method_func(X)
    D = np.diag(np.mean(D, axis=0))
    ind = np.argsort(D)[::-1]
    D = D[ind]
    V = V[:, ind]
    return V, D


def gw_csp_kernel(
    X: ndarray, y: ndarray, ajd_method: str = "uwedge"
) -> Tuple[ndarray, ndarray, ndarray, ndarray]:
    """Grosse-Wentrup AJD method based on paper [1]_.

    Parameters
    ----------
    X : ndarray
        eeg data, shape (n_trials, n_channels, n_samples).
    y : ndarray
        labels, shape (n_trials).
    ajd_method : str, optional
        ajd methods, 'uwedge' 'rjd' and 'ajd_pham', by default 'uwedge'.

    Returns
    -------
    W: ndarray
        Spatial filters, shape (n_channels, n_filters).
    D: ndarray
        Eigenvalues of spatial filters, shape (n_filters,).
    A: ndarray
        Spatial patterns, shape (n_channels, n_patterns).
    mutual_info: ndarray
        Mutual informaiton values, shape (n_filters).

    References
    ----------
    .. [1] Grosse-Wentrup, Moritz, and Martin Buss. "Multiclass common spatial patterns and information theoretic feature
           extraction." Biomedical Engineering, IEEE Transactions on 55, no. 8 (2008): 1991-2000.
    """
    X, y = np.copy(X), np.copy(y)
    labels = np.unique(y)
    X = X - np.mean(X, axis=-1, keepdims=True)

    Cx_list = []
    for label in labels:
        C = covariances(X[y == label])
        # trace normalization
        C = C / np.trace(C, axis1=-1, axis2=-2)[:, np.newaxis, np.newaxis]
        Cx_list.append(np.mean(C, axis=0))
    Cx = np.stack(Cx_list)
    W, D = ajd(Cx, method=ajd_method)
    # Ctot = np.mean(Cx, axis=0)
    # W = W / np.sqrt(np.diag(W.T@Ctot@W))
    W = W / np.sqrt(D)

    # compute mutual information values
    Pc = [np.mean(y == label) for label in labels]
    mutual_info_list = []
    for j in range(W.shape[-1]):
        a = 0
        b = 0
        for i in range(len(labels)):
            # tmp = np.dot(np.dot(W[j], self.C_[i]), W[j].T)
            tmp = W[:, j].T @ Cx[i] @ W[:, j]
            a += Pc[i] * np.log(np.sqrt(tmp))
            b += Pc[i] * (tmp**2 - 1)
        mi = -(a + (3.0 / 16) * (b**2))
        mutual_info_list.append(mi)
    mutual_info = np.array(mutual_info_list)
    ix = np.argsort(mutual_info)[::-1]
    W = W[:, ix]
    mutual_info = mutual_info[ix]
    D = D[ix]
    A = robust_pattern(W, Cx[0], W.T @ Cx[0] @ W)
    return W, D, A, mutual_info


class CSP(BaseEstimator, TransformerMixin):
    """Common Spatial Pattern.

    if n_components is None, auto finding the best number of components with gridsearch. The upper searching limit is
    determined by max_components, default is half of the number of channels.
    """

    def __init__(
        self, n_components: Optional[int] = None, max_components: Optional[int] = None
    ):
        self.n_components = n_components
        self.max_components = max_components

    def fit(self, X: ndarray, y: ndarray):
        self.classes_ = np.unique(y)
        self.W_, self.D_, self.A_ = csp_kernel(X, y)
        # resorting with 0.5 threshold
        self.D_ = np.abs(self.D_ - 0.5)
        ind = np.argsort(self.D_, axis=-1)[::-1]
        self.W_, self.D_, self.A_ = self.W_[:, ind], self.D_[ind], self.A_[:, ind]

        # auto-tuning
        if self.n_components is None:
            estimator = make_pipeline(*[CSP(n_components=self.n_components), SVC()])
            if self.max_components is None:
                params = {"csp__n_components": np.arange(1, self.W_.shape[1] + 1)}
            else:
                params = {"csp__n_components": np.arange(1, self.max_components + 1)}

            n_splits = np.min(np.unique(y, return_counts=True)[1])
            n_splits = 5 if n_splits > 5 else n_splits

            gs = GridSearchCV(
                estimator,
                param_grid=params,
                scoring="accuracy",
                cv=StratifiedKFold(n_splits=n_splits, shuffle=True),
                refit=False,
                n_jobs=-1,
                verbose=False,
            )
            gs.fit(X, y)
            self.best_n_components_ = gs.best_params_["csp__n_components"]
        return self

    def transform(self, X: ndarray):
        n_components = (
            self.best_n_components_ if self.n_components is None else self.n_components
        )
        return csp_feature(self.W_, X, n_components=n_components)


class MultiCSP(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        n_components: Optional[int] = None,
        max_components: Optional[int] = None,
        multiclass: str = "ovr",
        ajd_method: str = "uwedge",
    ):
        self.n_components = n_components
        self.max_components = max_components
        self.multiclass = multiclass
        self.ajd_method = ajd_method

    def fit(self, X: ndarray, y: ndarray):
        self.classes_ = np.unique(y)

        if self.multiclass == "ovr":
            self.estimator_ = OneVsRestClassifier(
                make_pipeline(
                    *[
                        CSP(
                            n_components=self.n_components,
                            max_components=self.max_components,
                        ),
                        SVC(),
                    ]
                ),
                n_jobs=-1,
            )
            self.estimator_.fit(X, y)

        elif self.multiclass == "ovo":
            self.estimator_ = OneVsOneClassifier(
                make_pipeline(
                    *[
                        CSP(
                            n_components=self.n_components,
                            max_components=self.max_components,
                        ),
                        SVC(),
                    ]
                ),
                n_jobs=-1,
            )
            # patching avoiding 2d array check
            self.estimator_._validate_data = partial(
                self.estimator_._validate_data, allow_nd=True
            )
            self.estimator_.fit(X, y)

        elif self.multiclass == "grosse-wentrup":
            self.W_, _, self.A_, self.mutualinfo_values_ = gw_csp_kernel(
                X, y, ajd_method=self.ajd_method
            )
            if self.n_components is None:
                estimator = make_pipeline(
                    *[
                        MultiCSP(
                            n_components=self.n_components,
                            multiclass="grosse-wentrup",
                            ajd_method=self.ajd_method,
                        ),
                        SVC(),
                    ]
                )
                if self.max_components is None:
                    params = {
                        "multicsp__n_components": np.arange(1, self.W_.shape[1] + 1)
                    }
                else:
                    params = {
                        "multicsp__n_components": np.arange(1, self.max_components + 1)
                    }

                n_splits = np.min(np.unique(y, return_counts=True)[1])
                n_splits = 5 if n_splits > 5 else n_splits
                gs = GridSearchCV(
                    estimator,
                    param_grid=params,
                    scoring="accuracy",
                    cv=StratifiedKFold(n_splits=n_splits, shuffle=True),
                    refit=False,
                    n_jobs=-1,
                    verbose=False,
                )
                gs.fit(X, y)
                self.best_n_components_ = gs.best_params_["multicsp__n_components"]
        else:
            raise ValueError("not a valid multiclass strategy")
        return self

    def transform(self, X: ndarray):
        if self.multiclass == "grosse-wentrup":
            n_components = (
                self.best_n_components_
                if self.n_components is None
                else self.n_components
            )
            features = csp_feature(self.W_, X, n_components=n_components)
        else:
            features = np.concatenate(
                [est[0].transform(X) for est in self.estimator_.estimators_], axis=-1
            )
        return features


def spoc_kernel(X: ndarray, y: ndarray) -> Tuple[ndarray, ndarray, ndarray]:
    """Source Power Comodulation (SPoC) based on paper [1]_.

    It is a continous CSP-like method.

    Parameters
    ----------
    X : ndarray
        eeg data, shape (n_trials, n_channels, n_samples)
    y : ndarray
        labels, shape (n_trials)

    Returns
    -------
    W: ndarray
        Spatial filters, shape (n_channels, n_filters).
    D: ndarray
        Eigenvalues of spatial filters, shape (n_filters,).
    A: ndarray
        Spatial patterns, shape (n_channels, n_patterns).

    References
    ----------
    .. [1] Sven Dähne, Frank C. Meinecke, Stefan Haufe, Johannes Höhne, Michael Tangermann, Klaus-Robert Müller,
           and Vadim V. Nikulin. SPoC: a novel framework for relating the amplitude of neuronal oscillations to
           behaviorally relevant parameters. NeuroImage, 86:111–122, 2014. doi:10.1016/j.neuroimage.2013.07.079.
    """
    X, weights = np.copy(X), np.copy(y)
    # eps = np.finfo(X.dtype).eps
    X = X - np.mean(X, axis=-1, keepdims=True)
    weights = weights - np.mean(weights)
    weights = weights / np.std(weights)
    Cx = covariances(X)
    # trace normalization
    Cx = Cx / np.trace(Cx, axis1=-1, axis2=-2)[:, np.newaxis, np.newaxis]
    C = np.mean(Cx, axis=0)
    Cz = np.mean(weights[:, np.newaxis, np.newaxis] * Cx, axis=0)

    # check positive-definiteness
    C = nearestPD(C)
    Cz = nearestPD(Cz)

    # TODO: direct copy from pyriemann, need verify
    D, W = eigh(Cz, C)
    ind = np.argsort(D)[::-1]
    D = D[ind]
    W = W[:, ind]

    A = robust_pattern(W, Cz, W.T @ Cz @ W)
    return W, D, A


class SPoC(BaseEstimator, TransformerMixin):
    """Source Power Comodulation (SPoC).

    For continuous data, not verified.

    """

    def __init__(
        self, n_components: Optional[int] = None, max_components: Optional[int] = None
    ):
        self.n_components = n_components
        self.max_components = max_components

    def fit(self, X: ndarray, y: ndarray):
        self.W_, self.D_, self.A_ = spoc_kernel(X, y)

        # auto-tuning
        if self.n_components is None:
            estimator = make_pipeline(
                *[SPoC(n_components=self.n_components), Ridge(alpha=0.5)]
            )
            if self.max_components is None:
                params = {"spoc__n_components": np.arange(1, self.W_.shape[1] + 1)}
            else:
                params = {"spoc__n_components": np.arange(1, self.max_components + 1)}

            test_size = 0.2 if len(y) > 5 else 1 / len(y)

            gs = GridSearchCV(
                estimator,
                param_grid=params,
                scoring="neg_root_mean_squared_error",
                cv=ShuffleSplit(n_splits=5, test_size=test_size),
                refit=False,
                n_jobs=-1,
                verbose=False,
            )
            gs.fit(X, y)
            self.best_n_components_ = gs.best_params_["spoc__n_components"]

    def transform(self, X: ndarray):
        n_components = (
            self.best_n_components_ if self.n_components is None else self.n_components
        )
        return csp_feature(self.W_, X, n_components=n_components)


class FBCSP(FilterBank):
    """FBCSP.

    FilterBank CSP based on paper [1]_.

    References
    ----------
    .. [1] Ang K K, Chin Z Y, Zhang H, et al. Filter bank common spatial pattern (FBCSP) in brain-computer interface[C]//2008
           IEEE International Joint Conference on Neural Networks (IEEE World Congress on Computational Intelligence). IEEE,
           2008: 2390-2397.
    """

    def __init__(
        self,
        n_components: Optional[int] = None,
        max_components: Optional[int] = None,
        n_mutualinfo_components: Optional[int] = None,
        filterbank: List[ndarray] = [],
    ):
        self.n_components = n_components
        self.max_components = max_components
        self.n_mutualinfo_components = n_mutualinfo_components
        self.filterbank = filterbank
        super().__init__(
            CSP(n_components=n_components, max_components=max_components),
            filterbank=filterbank,
        )

    def fit(self, X: ndarray, y: ndarray):  # type: ignore[override]
        super().fit(X, y)
        features = super().transform(X)
        if self.n_mutualinfo_components is None:
            estimator = make_pipeline(
                *[SelectKBest(score_func=mutual_info_classif, k="all"), SVC()]
            )
            params = {"selectkbest__k": np.arange(1, features.shape[1] + 1)}

            n_splits = np.min(np.unique(y, return_counts=True)[1])
            n_splits = 5 if n_splits > 5 else n_splits

            gs = GridSearchCV(
                estimator,
                param_grid=params,
                scoring="accuracy",
                cv=StratifiedKFold(n_splits=n_splits, shuffle=True),
                refit=False,
                n_jobs=-1,
                verbose=False,
            )
            gs.fit(features, y)
            self.best_n_mutualinfo_components_ = gs.best_params_["selectkbest__k"]
            self.selector_ = SelectKBest(
                score_func=mutual_info_classif, k=self.best_n_mutualinfo_components_
            )
        else:
            self.selector_ = SelectKBest(
                score_func=mutual_info_classif, k=self.n_mutualinfo_components
            )
        self.selector_.fit(features, y)
        return self

    def transform(self, X: ndarray):  # type: ignore[override]
        features = super().transform(X)
        features = self.selector_.transform(features)
        return features


class FBMultiCSP(FilterBank):
    def __init__(
        self,
        n_components: Optional[int] = None,
        max_components: Optional[int] = None,
        multiclass: str = "ovr",
        ajd_method: str = "uwedge",
        n_mutualinfo_components: Optional[int] = None,
        filterbank: List[ndarray] = [],
    ):
        self.n_components = n_components
        self.max_components = max_components
        self.multiclass = multiclass
        self.ajd_method = ajd_method
        self.n_mutualinfo_components = n_mutualinfo_components
        self.filterbank = filterbank
        self.n_mutualinfo_components = n_mutualinfo_components
        super().__init__(
            MultiCSP(
                n_components=n_components,
                max_components=max_components,
                multiclass=multiclass,
                ajd_method=ajd_method,
            ),
            filterbank=filterbank,
        )

    def fit(self, X: ndarray, y: ndarray):  # type: ignore[override]
        super().fit(X, y)
        features = super().transform(X)
        if self.n_mutualinfo_components is None:
            estimator = make_pipeline(
                *[SelectKBest(score_func=mutual_info_classif, k="all"), SVC()]
            )
            params = {"selectkbest__k": np.arange(1, features.shape[1] + 1)}

            n_splits = np.min(np.unique(y, return_counts=True)[1])
            n_splits = 5 if n_splits > 5 else n_splits

            gs = GridSearchCV(
                estimator,
                param_grid=params,
                scoring="accuracy",
                cv=StratifiedKFold(n_splits=n_splits, shuffle=True),
                refit=False,
                n_jobs=-1,
                verbose=False,
            )
            gs.fit(features, y)
            self.best_n_mutualinfo_components_ = gs.best_params_["selectkbest__k"]
            self.selector_ = SelectKBest(
                score_func=mutual_info_classif, k=self.best_n_mutualinfo_components_
            )
        else:
            self.selector_ = SelectKBest(
                score_func=mutual_info_classif, k=self.n_mutualinfo_components
            )
        self.selector_.fit(features, y)
        return self

    def transform(self, X: ndarray):  # type: ignore[override]
        features = super().transform(X)
        features = self.selector_.transform(features)
        return features



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/01/07
# License: MIT License
"""
Zhou2016.
"""
import os
import zipfile
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

from mne.io import read_raw_cnt, Raw
from mne.channels import make_standard_montage
from .base import BaseDataset
from ..utils.download import mne_data_path
from ..utils.channels import upper_ch_names

ZHOU_URL = "https://ndownloader.figshare.com/files/3662952"


class Zhou2016(BaseDataset):
    """Motor Imagery dataset from Zhou et al 2016.

    Dataset from the article *A Fully Automated Trial Selection Method for
    Optimization of Motor Imagery Based Brain-Computer Interface* [1]_.
    This dataset contains data recorded on 4 subjects performing 3 type of
    motor imagery: left hand, right hand and feet.

    Every subject went through three sessions, each of which contained two
    consecutive runs with several minutes inter-run breaks, and each run
    comprised 75 trials (25 trials per class). The intervals between two
    sessions varied from several days to several months.

    A trial started by a short beep indicating 1 s preparation time,
    and followed by a red arrow pointing randomly to three directions (left,
    right, or bottom) lasting for 5 s and then presented a black screen for
    4 s. The subject was instructed to immediately perform the imagination
    tasks of the left hand, right hand or foot movement respectively according
    to the cue direction, and try to relax during the black screen.

    References
    ----------

    .. [1] Zhou B, Wu X, Lv Z, Zhang L, Guo X (2016) A Fully Automated
           Trial Selection Method for Optimization of Motor Imagery Based
           Brain-Computer Interface. PLoS ONE 11(9).
           https://doi.org/10.1371/journal.pone.0162657
    """

    _EVENTS = {"left_hand": (1, (0, 5)), "right_hand": (2, (0, 5)), "feet": (3, (0, 5))}

    _CHANNELS = [
        "FP1",
        "FP2",
        "FC3",
        "FCZ",
        "FC4",
        "C3",
        "CZ",
        "C4",
        "CP3",
        "CPZ",
        "CP4",
        "O1",
        "OZ",
        "O2",
    ]

    def __init__(self):
        super().__init__(
            dataset_code="zhou2016",
            subjects=list(range(1, 5)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=250,
            paradigm="imagery",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))
        subject = cast(int, subject)
        url = "{:s}".format(ZHOU_URL)
        file_dest = mne_data_path(
            url,
            self.dataset_code,
            path=path,
            proxies=proxies,
            force_update=force_update,
            update_path=update_path,
        )
        parent_dir = Path(file_dest).parent

        if not os.path.exists(os.path.join(parent_dir, "data")):
            # decompression the data
            with zipfile.ZipFile(file_dest, "r") as archive:
                archive.extractall(path=parent_dir)
        dests: List[List[Union[str, Path]]] = []
        for session in range(1, 4):
            runs: List[Union[str, Path]] = []
            for run in ["A", "B"]:
                runs.append(
                    os.path.join(
                        parent_dir,
                        "data",
                        "S{:d}_{:d}{:s}.cnt".format(subject, session, run),
                    )
                )
            dests.append(runs)
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        sess = dict()
        for isess, run_dests in enumerate(dests):
            runs = dict()
            for irun, run_file in enumerate(run_dests):
                raw = read_raw_cnt(run_file, eog=["VEOU", "VEOL"], preload=True)
                raw = upper_ch_names(raw)
                raw.set_montage(montage)

                runs["run_{:d}".format(irun)] = raw
            sess["session_{:d}".format(isess)] = runs
        return sess



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2020/12/30
# License: MIT License
"""
Physionet MI.
"""
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

import numpy as np
from mne.io import Raw, read_raw_edf
from mne.channels import make_standard_montage

from .base import BaseDataset
from ..utils.download import mne_data_path
from ..utils.channels import upper_ch_names

PHYSIONET_URL = "http://www.physionet.org/pn4/eegmmidb/"


class BasePhysionet(BaseDataset):
    """Physionet Motor Imagery dataset.

    Physionet MI dataset: https://physionet.org/pn4/eegmmidb/

    This data set consists of over 1500 one- and two-minute EEG recordings,
    obtained from 109 volunteers.

    Subjects performed different motor/imagery tasks while 64-channel EEG were
    recorded using the BCI2000 system (http://www.bci2000.org).
    Each subject performed 14 experimental runs: two one-minute baseline runs
    (one with eyes open, one with eyes closed), and three two-minute runs of
    each of the four following tasks:

    1. A target appears on either the left or the right side of the screen.
       The subject opens and closes the corresponding fist until the target
       disappears. Then the subject relaxes.

    2. A target appears on either the left or the right side of the screen.
       The subject imagines opening and closing the corresponding fist until
       the target disappears. Then the subject relaxes.

    3. A target appears on either the top or the bottom of the screen.
       The subject opens and closes either both fists (if the target is on top)
       or both feet (if the target is on the bottom) until the target
       disappears. Then the subject relaxes.

    4. A target appears on either the top or the bottom of the screen.
       The subject imagines opening and closing either both fists
       (if the target is on top) or both feet (if the target is on the bottom)
       until the target disappears. Then the subject relaxes.

    parameters
    ----------

    imagined: bool (default True)
        if True, return runs corresponding to motor imagination.

    executed: bool (default False)
        if True, return runs corresponding to motor execution.

    references
    ----------

    .. [1] Schalk, G., McFarland, D.J., Hinterberger, T., Birbaumer, N. and
           Wolpaw, J.R., 2004. BCI2000: a general-purpose brain-computer
           interface (BCI) system. IEEE Transactions on biomedical engineering,
           51(6), pp.1034-1043.

    .. [2] Goldberger, A.L., Amaral, L.A., Glass, L., Hausdorff, J.M., Ivanov,
           P.C., Mark, R.G., Mietus, J.E., Moody, G.B., Peng, C.K., Stanley,
           H.E. and PhysioBank, P., PhysioNet: components of a new research
           resource for complex physiologic signals Circulation 2000 Volume
           101 Issue 23 pp. E215–E220.
    """

    _EVENTS = {
        "rest": (1, (0, 3)),
        "left_hand": (2, (0, 3)),
        "right_hand": (3, (0, 3)),
        "hands": (4, (0, 3)),
        "feet": (5, (0, 3)),
    }

    _CHANNELS = [
        "FC5",
        "FC3",
        "FC1",
        "FCZ",
        "FC2",
        "FC4",
        "FC6",
        "C5",
        "C3",
        "C1",
        "CZ",
        "C2",
        "C4",
        "C6",
        "CP5",
        "CP3",
        "CP1",
        "CPZ",
        "CP2",
        "CP4",
        "CP6",
        "FP1",
        "FPZ",
        "FP2",
        "AF7",
        "AF3",
        "AFZ",
        "AF4",
        "AF8",
        "F7",
        "F5",
        "F3",
        "F1",
        "FZ",
        "F2",
        "F4",
        "F6",
        "F8",
        "FT7",
        "FT8",
        "T7",
        "T8",
        "T9",
        "T10",
        "TP7",
        "TP8",
        "P7",
        "P5",
        "P3",
        "P1",
        "PZ",
        "P2",
        "P4",
        "P6",
        "P8",
        "PO7",
        "PO3",
        "POZ",
        "PO4",
        "PO8",
        "O1",
        "OZ",
        "O2",
        "IZ",
    ]

    def __init__(self, paradigm: str, is_imagined: bool = True):
        super().__init__(
            dataset_code="eegbci",
            subjects=list(range(1, 110)),
            events=self._EVENTS,  # type: ignore
            channels=self._CHANNELS,
            srate=160,
            paradigm=paradigm,
        )

        self.is_imagined = is_imagined
        self.baseline_runs = [1, 2]
        self.feet_runs = []
        self.hand_runs = []

        if self.is_imagined:
            self.feet_runs += [6, 10, 14]
            self.hand_runs += [4, 8, 12]
        else:
            self.feet_runs += [5, 9, 13]
            self.hand_runs += [3, 7, 11]

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        runs = self.baseline_runs + self.hand_runs + self.feet_runs

        dests = []
        for r in runs:
            base_url = "{u}S{s:03d}/S{s:03d}R{r:02d}.edf".format(
                u=PHYSIONET_URL, s=subject, r=r
            )
            dests.append(
                mne_data_path(
                    base_url,
                    self.dataset_code,
                    path=path,
                    proxies=proxies,
                    force_update=force_update,
                    update_path=update_path,
                )
            )
        return [dests]

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        sess = dict()
        for isess, run_dests in enumerate(dests):
            runs = dict()
            for irun, run_file in enumerate(run_dests):
                raw = read_raw_edf(run_file, preload=True)
                raw.rename_channels(lambda x: x.strip("."))
                raw = upper_ch_names(raw)
                raw.set_montage(montage)

                # change event id
                ori_desc = np.copy(raw.annotations.description)
                if irun == 0:
                    raw.annotations.description[ori_desc == "T0"] = 6
                    raw.annotations.description[ori_desc != "T0"] = 0
                if irun == 1:
                    raw.annotations.description[ori_desc == "T0"] = 7
                    raw.annotations.description[ori_desc != "T0"] = 0

                if irun in [2, 3, 4]:
                    raw.annotations.description[ori_desc == "T0"] = 1
                    raw.annotations.description[ori_desc == "T1"] = 2
                    raw.annotations.description[ori_desc == "T2"] = 3
                if irun in [5, 6, 7]:
                    raw.annotations.description[ori_desc == "T0"] = 1
                    raw.annotations.description[ori_desc == "T1"] = 4
                    raw.annotations.description[ori_desc == "T2"] = 5

                runs["run_{:d}".format(irun)] = raw
            sess["session_{:d}".format(isess)] = runs
        return sess

    def raw_hook(self, raw: Raw, caches: dict, verbose=None):
        # non-causal filtfilt
        raw.filter(3, 40, l_trans_bandwidth=2, h_trans_bandwidth=5, phase="zero-double")
        return raw, caches


class PhysionetMI(BasePhysionet):
    def __init__(self):
        super().__init__("imagery", is_imagined=True)


class PhysionetME(BasePhysionet):
    def __init__(self):
        super().__init__("movement", is_imagined=False)



from .base import BaseDataset
from .alex_mi import AlexMI
from .bnci import BNCI2014001, BNCI2014004
from .tsinghua import Wang2016, BETA
from .nakanishi2015 import Nakanishi2015
from .zhou2016 import Zhou2016
from .physionet import PhysionetMI, PhysionetME
from .cho2017 import Cho2017
from .munich2009 import MunichMI
from .schirrmeister2017 import Schirrmeister2017
from .tunerl import Weibo2014
from .cbcic import CBCIC2019001, CBCIC2019004



# -*- coding: utf-8 -*-
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2020/12/07
# License: MIT License
import mne
from mne.io import Raw
from typing import List, Union


def upper_ch_names(raw: Raw) -> Raw:
    """Uppercase all channel names in MNE Raw object.

    Parameters
    ----------
    raw : Raw
        MNE Raw object.

    Returns
    -------
    Raw
        MNE Raw object.
    """
    # raw.info['ch_names'] = [ch_name.upper() for ch_name in raw.info['ch_names']]
    # for i, ch in enumerate(raw.info['chs']):
    #     ch['ch_name'] = raw.info['ch_names'][i]
    raw = raw.rename_channels(
        {ch_name: ch_name.upper() for ch_name in raw.info["ch_names"]}
    )
    return raw


def pick_channels(
    ch_names: List[str],
    pick_chs: List[str],
    ordered: bool = True,
    match_case: Union[str, bool] = "auto",
) -> List[int]:
    """Wrapper of mne.pick_channels with match_case option.

    Parameters
    ----------
    ch_names : List[str]
        all channel names
    pick_chs : List[str]
        channel names to pick
    ordered : bool, optional
        if Ture, return picked channels in pick_chs order, by default True
    match_case : str, optional
        if True, pick channels in strict mode, by default 'auto'

    Returns
    -------
    List[int]
        indices of picked channels
    """

    """Wrapper of mne.pick_channels with match_case option.
    """
    if match_case == "auto":
        if len(set([ch_name.lower() for ch_name in ch_names])) < len(set(ch_names)):
            match_case = True
        else:
            match_case = False

    if match_case:
        picks = mne.pick_channels(ch_names, pick_chs, ordered=ordered)
    else:
        ch_names = [ch_name.lower() for ch_name in ch_names]
        pick_chs = [pick_ch.lower() for pick_ch in pick_chs]
        picks = mne.pick_channels(ch_names, pick_chs, ordered=ordered)

    return picks



# -*- coding: utf-8 -*-
# License: MIT License
"""
Logging system.

"""
import logging


def get_logger(log_name):
    """get system logger.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        Nonw
    Parameters
    ----------
    log_name: str,
        Name of logger.
    """
    logger = logging.getLogger(log_name)
    logger.setLevel(level=logging.INFO)

    handler = logging.FileHandler("log.txt", encoding="utf-8")
    handler.setLevel(level=logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def disable_log():
    """disable system logger.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        Nonw
    """
    logging.disable(logging.INFO)



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2020/12/28
# License: MIT License
"""
TUNERL Datasets

Weibo2014
"""
import os
import zipfile
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

import numpy as np
from mne import create_info
from mne.io import Raw, RawArray
from mne.channels import make_standard_montage
from .base import BaseDataset
from ..utils.download import mne_data_path
from ..utils.channels import upper_ch_names
from ..utils.io import loadmat

Weibo2014_URLs = [
    "https://dataverse.harvard.edu/api/access/datafile/2499178",
    "https://dataverse.harvard.edu/api/access/datafile/2499182",
    "https://dataverse.harvard.edu/api/access/datafile/2499179",
]


class Weibo2014(BaseDataset):
    """Motor Imagery dataset from Weibo et al 2014.

    Dataset from the article *Evaluation of EEG oscillatory patterns and
    cognitive process during simple and compound limb motor imagery* [1]_.

    It contains data recorded on 10 subjects, with 60 electrodes.

    This dataset was used to investigate the differences of the EEG patterns
    between simple limb motor imagery and compound limb motor
    imagery. Seven kinds of mental tasks have been designed, involving three
    tasks of simple limb motor imagery (left hand, right hand, feet), three
    tasks of compound limb motor imagery combining hand with hand/foot
    (both hands, left hand combined with right foot, right hand combined with
    left foot) and rest state.

    At the beginning of each trial (8 seconds), a white circle appeared at the
    center of the monitor. After 2 seconds, a red circle (preparation cue)
    appeared for 1 second to remind the subjects of paying attention to the
    character indication next. Then red circle disappeared and character
    indication (‘Left Hand’, ‘Left Hand & Right Foot’, et al) was presented on
    the screen for 4 seconds, during which the participants were asked to
    perform kinesthetic motor imagery rather than a visual type of imagery
    while avoiding any muscle movement. After 7 seconds, ‘Rest’ was presented
    for 1 second before next trial (Fig. 1(a)). The experiments were divided
    into 9 sections, involving 8 sections consisting of 60 trials each for six
    kinds of MI tasks (10 trials for each MI task in one section) and one
    section consisting of 80 trials for rest state. The sequence of six MI
    tasks was randomized. Intersection break was about 5 to 10 minutes.

    References
    -----------
    .. [1] Yi, Weibo, et al. "Evaluation of EEG oscillatory patterns and
           cognitive process during simple and compound limb motor imagery."
           PloS one 9.12 (2014). https://doi.org/10.1371/journal.pone.0114853
    """

    _EVENTS = {
        "left_hand": (1, (3, 7)),
        "right_hand": (2, (3, 7)),
        "hands": (3, (3, 7)),
        "feet": (4, (3, 7)),
        "left_hand_right_foot": (5, (3, 7)),
        "right_hand_left_foot": (6, (3, 7)),
        "rest": (7, (3, 7)),
    }

    _CHANNELS = [
        "FP1",
        "FPZ",
        "FP2",
        "AF3",
        "AF4",
        "F7",
        "F5",
        "F3",
        "F1",
        "Fz",
        "F2",
        "F4",
        "F6",
        "F8",
        "FT7",
        "FC5",
        "FC3",
        "FC1",
        "FCz",
        "FC2",
        "FC4",
        "FC6",
        "FT8",
        "T7",
        "C5",
        "C3",
        "C1",
        "CZ",
        "C2",
        "C4",
        "C6",
        "T8",
        "TP7",
        "CP5",
        "CP3",
        "CP1",
        "CPZ",
        "CP2",
        "CP4",
        "CP6",
        "TP8",
        "P7",
        "P5",
        "P3",
        "P1",
        "Pz",
        "P2",
        "P4",
        "P6",
        "P8",
        "PO7",
        "PO5",
        "PO3",
        "POZ",
        "PO4",
        "PO6",
        "PO8",
        "O1",
        "OZ",
        "O2",
    ]

    def __init__(self):
        super().__init__(
            dataset_code="weibo2014",
            subjects=list(range(1, 11)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=200,
            paradigm="imagery",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:

        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        if subject in range(1, 5):
            sub_names = ["cl", "cyy", "kyf", "lnn"]
            inc = 0
            file_dest = mne_data_path(
                Weibo2014_URLs[0],
                "tunerl",
                path=path,
                proxies=proxies,
                force_update=force_update,
                update_path=update_path,
            )
        elif subject in range(5, 8):
            sub_names = ["ls", "ry", "wcf"]
            inc = 4
            file_dest = mne_data_path(
                Weibo2014_URLs[1],
                "tunerl",
                path=path,
                proxies=proxies,
                force_update=force_update,
                update_path=update_path,
            )
        else:
            sub_names = ["wx", "yyx", "zd"]
            inc = 7
            file_dest = mne_data_path(
                Weibo2014_URLs[2],
                "tunerl",
                path=path,
                proxies=proxies,
                force_update=force_update,
                update_path=update_path,
            )

        parent_dir = Path(file_dest).parent

        if not os.path.exists(
            os.path.join(parent_dir, "subject_{:d}.mat".format(subject))
        ):
            with zipfile.ZipFile(file_dest, "r") as archive:
                archive.extractall(path=parent_dir)
            for i, sub_name in enumerate(sub_names):
                os.rename(
                    os.path.join(parent_dir, "{}.mat".format(sub_name)),
                    os.path.join(parent_dir, "subject_{:d}.mat".format(i + inc + 1)),
                )

        dests: List[List[Union[str, Path]]] = [
            [os.path.join(parent_dir, "subject_{:d}.mat".format(subject))]
        ]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        sess = dict()
        for isess, run_dests in enumerate(dests):
            runs = dict()
            for irun, run_file in enumerate(run_dests):
                raw_mat = loadmat(run_file)
                epoch_data = raw_mat["data"] * 1e-6
                label = raw_mat["label"]

                stim = np.zeros((1, epoch_data.shape[1], epoch_data.shape[2]))
                stim[0, 0, :] = label

                data = np.concatenate((epoch_data, stim), axis=0)
                data = np.transpose(data, axes=(0, 2, 1))
                data = np.reshape(data, (data.shape[0], -1))

                ch_names = [ch_name.upper() for ch_name in self._CHANNELS] + [
                    "VEO",
                    "HEO",
                    "STI 014",
                ]
                ch_names.insert(57, "CB1")
                ch_names.insert(61, "CB2")
                ch_types = ["eeg"] * 62 + ["eog"] * 2
                ch_types[57] = "misc"
                ch_types[61] = "misc"
                ch_types = ch_types + ["stim"]

                info = create_info(
                    ch_names=ch_names, ch_types=ch_types, sfreq=self.srate
                )

                raw = RawArray(data=data, info=info)
                raw = upper_ch_names(raw)
                raw.set_montage(montage)
                runs["run_{:d}".format(irun)] = raw
            sess["session_{:d}".format(isess)] = runs
        return sess



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/1/07
# License: MIT License
from functools import partial
from typing import Union, Optional, Callable
import numpy as np
from numpy import ndarray
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.covariance import oas, ledoit_wolf, fast_mcd, empirical_covariance
from joblib import Parallel, delayed
from scipy.linalg import eigh

estimator = Callable[[ndarray], ndarray]


def isPD(B: ndarray) -> bool:
    """Returns true when input matrix is positive-definite, via Cholesky decompositon method.

    Parameters
    ----------
    B : ndarray
        Any matrix, shape (N, N)

    Returns
    -------
    bool
        True if B is positve-definite.

    Notes
    -----
        Use numpy.linalg rather than scipy.linalg. In this case, scipy.linalg has unpredictable behaviors.
    """

    try:
        _ = np.linalg.cholesky(B)
        return True
    except np.linalg.LinAlgError:
        return False


def nearestPD(A: ndarray) -> ndarray:
    """Find the nearest positive-definite matrix to input.

    Parameters
    ----------
    A : ndarray
        Any square matrxi, shape (N, N)

    Returns
    -------
    A3 : ndarray
        positive-definite matrix to A

    Notes
    -----
    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1]_, which
    origins at [2]_.

    References
    ----------
    .. [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd
    .. [2] N.J. Higham, "Computing a nearest symmetric positive semidefinite matrix" (1988):
           https://doi.org/10.1016/0024-3795(88)90223-6
    """

    B = (A + A.T) / 2
    _, s, V = np.linalg.svd(B)

    H = np.dot(V.T, np.dot(np.diag(s), V))

    A2 = (B + H) / 2

    A3 = (A2 + A2.T) / 2

    if isPD(A3):
        return A3

    print("Replace current matrix with the nearest positive-definite matrix.")

    spacing = np.spacing(np.linalg.norm(A))
    # The above is different from [1]. It appears that MATLAB's `chol` Cholesky
    # decomposition will accept matrixes with exactly 0-eigenvalue, whereas
    # Numpy's will not. So where [1] uses `eps(mineig)` (where `eps` is Matlab
    # for `numpy.spacing`), we use the above definition. CAVEAT: our `spacing`
    # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
    # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
    # `spacing` will, for Gaussian random matrixes of small dimension, be on
    # othe order of 1e-16. In practice, both ways converge, as the unit test
    # below suggests.
    eye = np.eye(A.shape[0])
    k = 1
    while not isPD(A3):
        mineig = np.min(np.real(np.linalg.eigvals(A3)))
        A3 += eye * (-mineig * k**2 + spacing)
        k += 1

    return A3


def _lwf(X: ndarray) -> ndarray:
    """Wrapper for sklearn ledoit wolf covariance estimator.

    Parameters
    ----------
    X : ndarray
        EEG signal, shape (n_channels, n_samples).

    Returns
    -------
    C : ndarray
        Estimated covariance, shape (n_channels, n_channels).
    """
    C, _ = ledoit_wolf(X.T)
    return C


def _oas(X: ndarray) -> ndarray:
    """Wrapper for sklearn oas covariance estimator.

    Parameters
    ----------
    X : ndarray
        EEG signal, shape (n_channels, n_samples).

    Returns
    -------
    C : ndarray
        Estimated covariance, shape (n_channels, n_channels).
    """
    C, _ = oas(X.T)
    return C


def _cov(X: ndarray) -> ndarray:
    """Wrapper for sklearn sample covariance estimator.

    Parameters
    ----------
    X : ndarray
        EEG signal, shape (n_channels, n_samples).

    Returns
    -------
    C : ndarray
        Estimated covariance, shape (n_channels, n_channels).
    """
    C = empirical_covariance(X.T)
    return C


def _mcd(X: ndarray) -> ndarray:
    """Wrapper for sklearn mcd covariance estimator.

    Parameters
    ----------
    X : ndarray
        EEG signal, shape (n_channels, n_samples).

    Returns
    -------
    C : ndarray
        Estimated covariance, shape (n_channels, n_channels).
    """
    _, C, _, _ = fast_mcd(X.T)
    return C


estimators = {
    "cov": _cov,
    "lwf": _lwf,
    "oas": _oas,
    "mcd": _mcd,
}


def _check_est(est: Union[str, estimator]) -> estimator:
    """Check if a given estimator is valid.

    Parameters
    ----------
    est : callable object or str
        Could be the name of estimator or a callable estimator itself.

    Returns
    -------
    est: callable object
        A callable estimator.
    """
    if callable(est):
        pass
    elif est in estimators.keys():
        est = estimators[est]
    else:
        raise ValueError(
            """%s is not an valid estimator ! Valid estimators are : %s or a
             callable function"""
            % (est, (" , ").join(estimators.keys()))
        )
    return est


def covariances(
    X: ndarray, estimator: Union[str, estimator] = "cov", n_jobs: int = 1
) -> ndarray:
    """Estimation of covariance matrix.

    Parameters
    ----------
    X : ndarray
        EEG signal, shape (..., n_channels, n_samples).
    estimator : str or callable object, optional
        Covariance estimator to use (the default is `cov`, which uses empirical covariance estimator). For regularization,
        consider `lwf` or `oas`.

        **supported estimators**

            `cov`: empirial covariance estimator

            `lwf`: ledoit wolf covariance estimator

            `oas`: oracle approximating shrinkage covariance estimator

            `mcd`: minimum covariance determinant covariance estimator
    n_jobs : int or None, optional
        The number of CPUs to use to do the computation (the default is 1, -1 for all processors).

    Returns
    -------
    covmats : ndarray
        covariance matrices, shape (..., n_channels, n_channels)

    See Also
    --------
    covariances_erp
    """
    X = np.asarray(X)
    X = np.atleast_2d(X)
    shape = X.shape
    X = np.reshape(X, (-1, shape[-2], shape[-1]))

    parallel = Parallel(n_jobs=n_jobs)
    est = _check_est(estimator)
    covmats = parallel(delayed(est)(x) for x in X)

    covmats = np.reshape(covmats, (*shape[:-2], shape[-2], shape[-2]))
    return covmats


class Covariance(BaseEstimator, TransformerMixin):
    """Estimation of covariance matrix.

    Parameters
    ----------
    estimator : str or callable object, optional
        Covariance estimator to use (the default is `cov`, which uses empirical covariance estimator). For regularization,
        consider `lwf` or `oas`.

        **supported estimators**

            `cov`: empirial covariance estimator

            `lwf`: ledoit wolf covariance estimator

            `oas`: oracle approximating shrinkage covariance estimator

            `mcd`: minimum covariance determinant covariance estimator
    n_jobs : int or None, optional
        The number of CPUs to use to do the computation (the default is 1, -1 for all processors).

    See Also
    --------
    ERPCovariance
    """

    def __init__(self, estimator="cov", n_jobs=1):
        self.estimator = estimator
        self.n_jobs = n_jobs

    def fit(self, X, y=None):
        """Not used, only for compatibility with sklearn API.

        Parameters
        ----------
        X : ndarray
            EEG signal, shape (..., n_channels, n_samples).
        y : ndarray
            Labels.

        Returns
        -------
        self : Covariance instance
            The Covariance instance.
        """
        return self

    def transform(self, X):
        """Transform EEG to covariance matrix.

        Parameters
        ----------
        X : ndarray
            EEG signal, shape (..., n_channels, n_samples).

        Returns
        -------
        covmats : ndarray
            Estimated covariances, shape (..., n_channels, n_channels)
        """
        covmats = covariances(X, estimator=self.estimator, n_jobs=self.n_jobs)
        return covmats


def matrix_operator(
    Ci: ndarray, operator: estimator, n_jobs: Optional[int] = None
) -> ndarray:
    """Apply operator to any matrix.

    Parameters
    ----------
    Ci : ndarray
        Input positive definite matrix.
    operator : callable object
        Operator function or callable object.
    n_jobs: int, optional
        the number of jobs to use.

    Returns
    -------
    Co : ndarray
        Operated matrix.

    Raises
    ------
    ValueError
        If Ci is not positive definite.

    Notes
    -----
    .. math::
        \mathbf{Ci} = \mathbf{V} \left( \mathbf{\Lambda} \\right) \mathbf{V}^T \\\\
        \mathbf{Co} = \mathbf{V} operator\left( \mathbf{\Lambda} \\right) \mathbf{V}^T

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{Ci}`.
    """

    def _single_matrix_operator(Ci: ndarray, operator: estimator) -> ndarray:
        eigvals, eigvects = eigh(Ci, check_finite=False)
        eigvals = np.diag(operator(eigvals))
        Co = eigvects @ eigvals @ eigvects.T
        return Co

    ori_shape = Ci.shape
    Ci = Ci.reshape((-1, *ori_shape[-2:]))
    Co = Parallel(n_jobs=n_jobs)(
        delayed(_single_matrix_operator)(C, operator) for C in Ci
    )
    Co = np.stack(Co)
    Co = Co.reshape((*ori_shape,))
    return Co


def sqrtm(Ci: ndarray, n_jobs: Optional[int] = None):
    """Return the matrix square root of a covariance matrix.

    Parameters
    ----------
    Ci : ndarray
        Input positive-definite matrix.

    Returns
    -------
    ndarray
        Square root matrix of Ci.

    Notes
    -----
    .. math::
        \mathbf{C} = \mathbf{V} \left( \mathbf{\Lambda} \\right)^{1/2} \mathbf{V}^T

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{Ci}`.
    """
    return matrix_operator(Ci, np.sqrt, n_jobs=n_jobs)


def logm(Ci: ndarray, n_jobs: Optional[int] = None):
    """Return the matrix logrithm of a covariance matrix.

    Parameters
    ----------
    Ci : ndarray
        Input positive-definite matrix.

    Returns
    -------
    ndarray
        Logrithm matrix of Ci.

    Notes
    -----
    .. math::
        \mathbf{C} = \mathbf{V} \log{(\mathbf{\Lambda})} \mathbf{V}^T

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{Ci}`.
    """
    return matrix_operator(Ci, np.log, n_jobs=n_jobs)


def expm(Ci: ndarray, n_jobs: Optional[int] = None):
    """Return the matrix exponential of a covariance matrix.

    Parameters
    ----------
    Ci : ndarray
        Input positive-definite matrix.

    Returns
    -------
    ndarray
        Exponential matrix of Ci.

    Notes
    -----
    .. math::
        \mathbf{C} = \mathbf{V} \exp{(\mathbf{\Lambda})} \mathbf{V}^T

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{Ci}`.
    """
    return matrix_operator(Ci, np.exp, n_jobs=n_jobs)


def invsqrtm(Ci: ndarray, n_jobs: Optional[int] = None):
    """Return the inverse matrix square root of a covariance matrix.

    Parameters
    ----------
    Ci : ndarray
        Input positive-definite matrix.

    Returns
    -------
    ndarray
        Inverse matrix square root of Ci.

    Notes
    -----
    .. math::
        \mathbf{C} = \mathbf{V} \left( \mathbf{\Lambda} \\right)^{-1/2} \mathbf{V}^T

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{Ci}`.
    """

    def isqrt(x):
        return 1.0 / np.sqrt(x)

    return matrix_operator(Ci, isqrt, n_jobs=n_jobs)


def powm(Ci: ndarray, alpha: float, n_jobs: Optional[int] = None):
    """Return the matrix power of a covariance matrix.

    Parameters
    ----------
    Ci : ndarray
        Input positive-definite matrix.
    alpha : float
        Exponent.

    Returns
    -------
    ndarray
        Power matrix of Ci.

    Notes
    -----
    .. math::
        \mathbf{C} = \mathbf{V} \left( \mathbf{\Lambda} \\right)^{\\alpha} \mathbf{V}^T

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{Ci}`.
    """
    power = partial(lambda x, alpha=None: x**alpha, alpha=alpha)
    return matrix_operator(Ci, power, n_jobs=n_jobs)



from .base_tmpl import BaseTmpl


def add(a, b):
    return a + b


def minus(a, b):
    return a - b


class TestDemo(BaseTmpl):

    def test_add(self):
        self.dbgPrint("3 == 1 + 2")
        self.assertEqual(3, add(1, 2))
        self.assertNotEqual(3, add(2, 2))

    def test_minus(self):
        self.assertEqual(1, minus(3, 2))



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2020/6/01
# License: MIT License
"""
Base Paradigm Design.

"""
from abc import ABCMeta, abstractmethod
from typing import Union, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import mne
from mne.utils import verbose
from joblib import Parallel, delayed

from ..utils import pick_channels
from ..datasets.base import BaseDataset


def label_encoder(y, labels):
    new_y = y.copy()
    for i, label in enumerate(labels):
        ix = y == label
        new_y[ix] = i
    return new_y


class BaseParadigm(metaclass=ABCMeta):
    """Abstract Base Paradigm."""

    def __init__(
        self,
        channels: Optional[List[str]] = None,
        events: Optional[List[str]] = None,
        intervals: Optional[List[Tuple[float, float]]] = None,
        srate: Optional[float] = None,
    ):
        """

        Parameters
        ----------
        channels : Optional[List[str]], optional
            selected channel names, if None use all channels in dataset, by default None
        events : Optional[List[str]], optional
            selected event names, if None use all events in dataset, by default None
        intervals : Optional[List[Tuple[Union[int, float]]]], optional
            selected event intervals, if None use default intervals in dataset.
            If only one interval passed, all events use the same interval.
            Otherwise the number of tuples should be the same as the number of events, by default None
        srate : Optional[float], optional
            sampling rate, if None use default srate in dataset, by default None
        """
        self.select_channels = (
            None if channels is None else [ch_name.upper() for ch_name in channels]
        )
        self.event_list = events
        self.intervals = intervals
        self.srate = srate
        self._raw_hook = None
        self._epochs_hook = None
        self._data_hook = None

    @abstractmethod
    def is_valid(self, dataset: BaseDataset) -> bool:
        """Verify the dataset is compatible with the paradigm.

        This method is called to verify dataset is compatible with the
        paradigm.

        This method should raise an error if the dataset is not compatible
        with the paradigm. This is for example the case if the
        dataset is an ERP dataset for motor imagery paradigm, or if the
        dataset does not contain any of the required events.

        Parameters
        ----------
        dataset : BaseDataset
            dataset
        """
        pass

    def _map_events_intervals(self, dataset: BaseDataset):
        event_list = self.event_list
        intervals = self.intervals

        if event_list is None:
            # use all events in dataset
            event_list = list(dataset.events.keys())

        used_events = {ev: dataset.events[ev][0] for ev in event_list}

        if intervals is None:
            used_intervals = {ev: dataset.events[ev][1] for ev in event_list}
        elif len(intervals) == 1:
            used_intervals = {ev: intervals[0] for ev in event_list}
        else:
            if len(event_list) != len(intervals):
                raise ValueError("intervals should be the same number of events")
            used_intervals = {
                ev: interval for ev, interval in zip(event_list, intervals)
            }

        return used_events, used_intervals

    def register_raw_hook(self, hook):
        """Register raw hook before epoch operation.

        Parameters
        ----------
        hook : callable object
            Callable object to process Raw object before epoch operation.
            Its' signature should look like:

            hook(raw, caches) -> raw, caches

            where caches is an dict stroing information, raw is MNE Raw instance.
        """
        self._raw_hook = hook

    def register_epochs_hook(self, hook):
        """Register epochs hook after epoch operation.

        Parameters
        ----------
        hook : callable object
            Callable object to process Epochs object after epoch operation.
            Its' signature should look like:

            hook(epochs, caches) -> epochs, caches

            where caches is an dict storing information, epochs is MNE Epochs instance.
        """
        self._epochs_hook = hook

    def register_data_hook(self, hook):
        """Register data hook before return data.

        Parameters
        ----------
        hook : callable object
            Callable object to process ndarray data before return it.
            Its' signature should look like:

            hook(X, y, meta, caches) -> X, y, meta, caches

            where caches is an dict storing information, X, y are ndarray object, meta is a pandas DataFrame instance.
        """
        self._data_hook = hook

    def unregister_raw_hook(self):
        """Unregister raw hook before epoch operation."""
        self._raw_hook = None

    def unregister_epochs_hook(self):
        """Register epochs hook after epoch operation."""
        self._epochs_hook = None

    def unregister_data_hook(self):
        """Register data hook before return data."""
        self._data_hook = None

    @verbose
    def _get_single_subject_data(self, dataset, subject_id, verbose=False):
        """Return data in micro-volt."""
        if not self.is_valid(dataset):
            raise TypeError(
                "Dataset {:s} is not valid for the current paradigm. Check your events and channels settings".format(
                    dataset.dataset_code
                )
            )

        # # events, interval checking
        used_events, used_intervals = self._map_events_intervals(dataset)

        Xs = {}
        ys = {}
        metas = {}

        data = dataset.get_data([subject_id])

        for subject, sessions in data.items():
            for session, runs in sessions.items():
                for run, raw in runs.items():
                    # do raw hook either self-implemented or dataset inherited
                    caches = {}
                    if self._raw_hook:
                        raw, caches = self._raw_hook(raw, caches)
                    elif hasattr(dataset, "raw_hook"):
                        raw, caches = dataset.raw_hook(raw, caches)

                    # pick selected channels by order
                    channels = (
                        dataset.channels
                        if self.select_channels is None
                        else self.select_channels
                    )
                    picks = pick_channels(raw.ch_names, channels, ordered=True)

                    # find available events, first check stim_channels then annotations
                    stim_channels = mne.utils._get_stim_channel(
                        None, raw.info, raise_error=False
                    )
                    if len(stim_channels) > 0:
                        events = mne.find_events(
                            raw, shortest_event=0, initial_event=True
                        )
                    else:
                        # convert event_id to its number type instead of default auto-renaming in 0.19.2
                        events, _ = mne.events_from_annotations(
                            raw, event_id=(lambda x: int(x))
                        )

                    for event_name in used_events.keys():
                        # mne.pick_events returns any matching events in include
                        # only raise Runtime Error when nothing is found
                        # then we just skip this event
                        try:
                            selected_events = mne.pick_events(
                                events, include=used_events[event_name]
                            )
                        except RuntimeError:
                            continue

                        # transform Raw to Epochs

                        epochs = mne.Epochs(
                            raw,
                            selected_events,
                            event_id={event_name: used_events[event_name]},
                            event_repeated="drop",
                            tmin=used_intervals[event_name][0],
                            tmax=used_intervals[event_name][1]
                            - 1.0 / raw.info["sfreq"],
                            picks=picks,
                            proj=False,
                            baseline=None,
                            preload=True,
                        )

                        # skip invalid time intervals
                        if len(epochs) == 0:
                            continue

                        # do epochs hook
                        if self._epochs_hook:
                            epochs, caches = self._epochs_hook(epochs, caches)
                        elif hasattr(dataset, "epochs_hook"):
                            epochs, caches = dataset.epochs_hook(epochs, caches)

                        # FIXME: is this resample reasonable?
                        if self.srate:
                            # as MNE suggested, decimate after extract epochs
                            # low-pass raw object in raw_hook to prevent aliasing problem
                            epochs = epochs.resample(self.srate)
                            # epochs = epochs.decimate(dataset.srate//self.srate)

                        # retrieve X, y and meta
                        X = epochs.get_data() * 1e6  # micro-volt default
                        y = epochs.events[:, -1]
                        trial_ids = np.argwhere(
                            events[:, -1] == list(epochs.event_id.values())[0]
                        ).reshape((-1))
                        meta = pd.DataFrame(
                            {
                                "subject": [subject] * len(epochs),
                                "session": [session] * len(epochs),
                                "run": [run] * len(epochs),
                                "event": [event_name] * len(epochs),
                                "trial_id": trial_ids,
                                "dataset": [dataset.dataset_code] * len(epochs),
                            }
                        )

                        # do data hook
                        if self._data_hook:
                            X, y, meta, caches = self._data_hook(X, y, meta, caches)
                        elif hasattr(dataset, "data_hook"):
                            X, y, meta, caches = dataset.data_hook(X, y, meta, caches)

                        # collecting data
                        pre_X = Xs.get(event_name)
                        if pre_X is not None:
                            Xs[event_name] = np.concatenate((pre_X, X), axis=0)
                        else:
                            Xs[event_name] = X

                        pre_y = ys.get(event_name)
                        if pre_y is not None:
                            ys[event_name] = np.concatenate((pre_y, y), axis=0)
                        else:
                            ys[event_name] = y

                        pre_meta = metas.get(event_name)
                        if pre_meta is not None:
                            metas[event_name] = pd.concat(
                                (pre_meta, meta), axis=0, ignore_index=True
                            )
                        else:
                            metas[event_name] = meta
        return Xs, ys, metas

    @verbose
    def get_data(
        self,
        dataset: BaseDataset,
        subjects: List[Union[int, str]] = [],
        label_encode: bool = True,
        return_concat: bool = False,
        n_jobs: int = -1,
        verbose: Optional[bool] = None,
    ) -> Tuple[
        Union[
            Dict[str, Union[np.ndarray, pd.DataFrame]], Union[np.ndarray, pd.DataFrame]
        ],
        ...,
    ]:
        """Get data from dataset with selected subjects.

        Parameters
        ----------
        dataset : BaseDataset
            dataset
        subjects : List[Union[int, str]],
            selected subjects, by default empty
        label_encode: bool, optional,
            if True, return y in label encode way
        return_concat : bool, optional
            if True, return concated ndarray object, otherwise return dict of events, by default False
        n_jobs : int, optional
            Parallel jobs, by default -1
        verbose : Optional[bool], optional
            verbose, by default None

        Returns
        -------
        Tuple[Union[Dict[str, Union[np.ndarray, pd.DataFrame]], Union[np.ndarray, pd.DataFrame]], ...]
            Xs, ys, metas, corresponding to data, label and meta data

        Raises
        ------
        TypeError
            raise error if dataset is not avaliable for the paradigm
        """
        if not self.is_valid(dataset):
            raise TypeError(
                "Dataset {:s} is not valid for the current paradigm. Check your events and channels settings".format(
                    dataset.dataset_code
                )
            )
        # events, interval checking
        used_events, used_intervals = self._map_events_intervals(dataset)

        Xs = {}
        ys = {}
        metas = {}

        X, y, meta = zip(
            *Parallel(n_jobs=n_jobs)(
                delayed(self._get_single_subject_data)(dataset, sub_id, verbose=verbose)
                for sub_id in subjects
            )
        )

        for event_name in used_events.keys():
            Xs[event_name] = np.concatenate(
                [X[i][event_name] for i in range(len(subjects)) if event_name in X[i]],
                axis=0,
            )
            ys[event_name] = np.concatenate(
                [y[i][event_name] for i in range(len(subjects)) if event_name in y[i]],
                axis=0,
            )
            metas[event_name] = pd.concat(
                [
                    meta[i][event_name]
                    for i in range(len(subjects))
                    if event_name in meta[i]
                ],
                axis=0,
                ignore_index=True,
            )

        if label_encode:
            event_list = list(used_events.keys())
            event_id = [dataset.events[e][0] for e in event_list]
            for event_name in used_events.keys():
                ys[event_name] = label_encoder(ys[event_name], event_id)

        # python gaurante values in insert order.
        if return_concat:
            Xs = np.concatenate(list(Xs.values()), axis=0)
            ys = np.concatenate(list(ys.values()), axis=0)
            metas = pd.concat(list(metas.values()), axis=0, ignore_index=True)

        return Xs, ys, metas

    def __str__(self):
        desc = "{}".format(self.__class__.__name__)
        return desc



# -*- coding:utf-8 -*-
"""
@ author: Jin Han
@ email: jinhan9165@gmail.com
@ Created on: 2022-05
version 1.0
update:
Refer: [1] Zhang, Yu, et al. "Spatial-temporal discriminant analysis for ERP-based brain-computer interface."
            IEEE Transactions on Neural Systems and Rehabilitation Engineering 21.2 (2013): 233-243.

Application: Spatial-Temporal Discriminant Analysis (STDA)

"""
import warnings

import numpy as np
from numpy import ndarray
from scipy import linalg as LA
from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin


def lda_kernel(X1: ndarray, X2: ndarray):
    """
    Linear Discriminant analysis kernel that is appliable to binary problems.

    Parameters
    ----------
    X1: ndarray of shape (n_samples, n_features)
        samples for class 1 (i.e. positive samples)

    X2: ndarray of shape (n_samples, n_features)
        samples for class 2 (i.e. negative samples)

    Returns
    -------
    weight_vec: ndarray of shape (1, n_features)
        weight vector.

    lda_threshold: float

    Note
    ----
    The test samples should be formatted as (n_samples, n_features).
        test sample is positive, if W @ test_sample.T > lda_thold.
        test sample is negative, if W @ test_sample.T <= lda_thold.
    """

    # mean feature vectors
    avg_feats1, avg_feats2 = X1.mean(axis=0, keepdims=True), X2.mean(
        axis=0, keepdims=True
    )

    # within-class scatter matrix
    X1_tmp, X2_tmp = (X1 - avg_feats1), (X2 - avg_feats2)
    Sw = X1_tmp.T @ X1_tmp + X2_tmp.T @ X2_tmp

    weight_vec = (LA.inv(Sw) @ (avg_feats1 - avg_feats2).T).T
    lda_threshold = weight_vec @ (avg_feats1.T + avg_feats2.T) / 2

    return weight_vec, lda_threshold.item()


def lda_proba(test_samples: ndarray, weight_vec: ndarray, lda_threshold: float):
    """Calculate decision value.

    Parameters
    ----------
    test_samples: 2-D, (n_samples, n_features)
    weight_vec: from LDA_kernel.
    lda_threshold: from LDA_kernel.

    Returns
    -------
    proba: ndarray of shape (n_samples,)
    """
    proba = weight_vec @ test_samples.T

    return proba.squeeze()


class STDA(BaseEstimator, TransformerMixin, ClassifierMixin):
    """
    Spatial-Temporal Discriminant Analysis (STDA).
    Note that the parameters naming are exactly the same as in the paper for convenient application.

    Parameters
    ----------
    L: int
        the number of eigenvectors retained for projection matrices.

    max_iter: int, default=400
        Max iteration times.

    eps: float, default=1e-5, also can be 1e-10.
        Error to guarantee convergence.
        Error = norm2(W(n) - W(n-1)), see more details in paper[1].

    Attributes
    ----------
    W1: ndarray of shape (D1, self.L)
        Weight vector. Actually, D1=n_chs.

    W2: ndarray of shape (D2, self.L)
        Weight vector. Actually, D2=n_features.

    iter_times: int
        Iteration times of STDA.

    wf: ndarray of shape (1, L*L)
        Weight vector of LDA after the raw features are projected by STDA.

    References
    ----------
    [1] Zhang, Yu, et al. "Spatial-temporal discriminant analysis for ERP-based brain-computer interface."
            IEEE Transactions on Neural Systems and Rehabilitation Engineering 21.2 (2013): 233-243.
    """

    def __init__(self, L: int = 6, max_iter: int = 400, eps: float = 1e-5):
        self.L = L
        self.max_iter = max_iter
        self.eps = eps

    def fit(self, X, y):
        """
        Fit Spatial-Temporal Discriminant Analysis (STDA) model.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_chs, n_features)
           Training data.

        y : array-like of shape (n_samples,)
            Target values. {-1, 1} or {0, 1}

        Returns
        -------
        self: object
              Fitted estimator (i.e. self.W1, self.W2).
        """
        self.classes_ = np.unique(y)
        _, n_chs, n_features = X.shape
        n_classes = len(self.classes_)

        if self.L > n_chs or self.L > n_features:
            raise ValueError(
                "The parameter L must be less than n_sample or n_features."
            )

        if n_classes != 2:
            raise NotImplementedError(
                "Only for binary problem. Multi-class STDA is not tested."
            )

        loc = [
            np.argwhere(y == self.classes_[idx_class]).squeeze()
            for idx_class in range(n_classes)
        ]
        X1, X2 = X[loc[0]], X[loc[1]]  # X1: negative samples. X2: positive samples.

        n_samples_c1, n_samples_c2 = X1.shape[0], X2.shape[0]
        # W1, W2 = np.ones((n_chs, self.L)), np.ones((n_features, self.L))
        W1, W2 = [], []
        W2.append(np.ones((n_features, n_features)))
        self.iter_times = 0
        while 1:
            self.iter_times += 1
            for k in range(1, n_classes + 1):
                if k == 1:
                    Y_mat_c1, Y_mat_c2 = np.matmul(X1, W2[-1]), np.matmul(X2, W2[-1])
                elif k == 2:
                    Y_mat_c1, Y_mat_c2 = np.matmul(W1[-1].T, X1).transpose(
                        (0, 2, 1)
                    ), np.matmul(W1[-1].T, X2).transpose((0, 2, 1))

                Y_bar_c1, Y_bar_c2 = Y_mat_c1.mean(axis=0), Y_mat_c2.mean(axis=0)
                Y_bar_all = ((Y_bar_c1 * n_samples_c1) + (Y_bar_c2 * n_samples_c2)) / (
                    n_samples_c1 + n_samples_c2
                )

                # construct Sb
                y1_tmp, y2_tmp = Y_bar_c1 - Y_bar_all, Y_bar_c2 - Y_bar_all
                Sb = n_samples_c1 * (y1_tmp @ y1_tmp.T) + n_samples_c2 * (
                    y2_tmp @ y2_tmp.T
                )

                # construct Sw
                Y_mat_c1 -= Y_bar_all
                Y_mat_c2 -= Y_bar_all
                Sw = (
                    np.matmul(Y_mat_c1, Y_mat_c1.transpose((0, 2, 1)))
                    + np.matmul(Y_mat_c2, Y_mat_c2.transpose((0, 2, 1)))
                ).sum(axis=0)

                # eig_vals, eig_vecs = LA.eig(np.dot(LA.inv(Sw), Sb))
                eig_vals, eig_vecs = LA.eigh(Sb, Sw)
                loc_idx = eig_vals.argsort()[::-1]  # descending order

                eig_vals = eig_vals[loc_idx]

                if k == 1:
                    W1.append(
                        eig_vecs[:, loc_idx][:, : self.L]
                    )  # return indices in ascending order and reverse
                else:
                    W2.append(eig_vecs[:, loc_idx][:, : self.L])

            # stop criterion
            if self.iter_times >= 4:
                if (
                    LA.norm(W1[-1] - W1[-2], ord=2) < self.eps
                    and LA.norm(W2[-1] - W2[-2], ord=2) < self.eps
                ) or (self.iter_times > self.max_iter):
                    self.W1, self.W2 = W1[-1], W2[-1]

                    f_c1 = np.matmul(np.matmul(self.W1.T, X1), self.W2).reshape(
                        n_samples_c1, -1
                    )
                    f_c2 = np.matmul(np.matmul(self.W1.T, X2), self.W2).reshape(
                        n_samples_c2, -1
                    )

                    self.wf, _ = lda_kernel(f_c2, f_c1)

                    return self

                if self.iter_times > 200:
                    warnings.warn(
                        "The alternating iteration has been performed many times (>200). "
                        "Model may be un-convergence."
                    )

    def transform(self, Xtest):
        """Project data and Get the decision values.

        Parameters
        ----------
        Xtest: ndarray of shape (n_samples, n_features).
            Input test data.

        Returns
        -------
        H_dv: ndarray of shape (n_samples, )
            decision values.
        """
        n_samples = Xtest.shape[0]
        f_hat = np.matmul(np.matmul(self.W1.T, Xtest), self.W2).reshape(n_samples, -1)

        H_dv = f_hat @ self.wf.T

        return H_dv.squeeze()


if __name__ == "__main__":
    clf_stda = STDA(L=6)

    # X = np.random.randn(1080*2, 16, 19)
    X = np.random.randint(-100, 100, (1080 * 2, 16, 19))
    y = np.hstack((np.ones(1080, dtype=int), np.ones(1080, dtype=int) * -1))

    clf_stda.fit(X, y)



"""
Deep4Net.
Modified from https://github.com/braindecode/braindecode/blob/master/braindecode/models/deep4.py
 -author: Xie YT
 -Created on: 2022-07-02
 -update log: ...

"""


import numpy as np
from torch import nn
from torch.nn import init
from torch.nn.functional import elu

from .base import (
    Expression,
    AvgPool2dWithConv,
    Ensure4d,
    identity,
    transpose_time_to_spat,
    squeeze_final_output,
    np_to_th,
    SkorchNet,
)


@SkorchNet
class Deep4Net(nn.Sequential):
    """Deep ConvNet model from Schirrmeister et al 2017.

    Model described in [Schirrmeister2017]_.

    Parameters
    ----------
    n_channels : int
        XXX

    References
    ----------
    .. [Schirrmeister2017] Schirrmeister, R. T., Springenberg, J. T., Fiederer,
       L. D. J., Glasstetter, M., Eggensperger, K., Tangermann, M., Hutter, F.
       & Ball, T. (2017).
       Deep learning with convolutional neural networks for EEG decoding and
       visualization.
       Human Brain Mapping , Aug. 2017.
       Online: http://dx.doi.org/10.1002/hbm.23730

    -author: Xie YT
    -Created on: 2022-07-02
    -update log: ...

    """

    def __init__(
        self,
        n_channels,
        n_samples,
        n_classes,
    ):
        super().__init__()
        final_conv_length = "auto"
        n_filters_time = 25
        n_filters_spat = 25
        filter_time_length = 10
        pool_time_length = 3
        pool_time_stride = 3
        n_filters_2 = 50
        filter_length_2 = 10
        n_filters_3 = 100
        filter_length_3 = 10
        n_filters_4 = 200
        filter_length_4 = 10
        first_nonlin = elu
        first_pool_mode = "max"
        first_pool_nonlin = identity
        later_nonlin = elu
        later_pool_mode = "max"
        later_pool_nonlin = identity
        drop_prob = 0.5
        # double_time_convs = False
        split_first_layer = True
        batch_norm = True
        batch_norm_alpha = 0.1
        stride_before_pool = False

        if final_conv_length == "auto":
            assert n_samples is not None
        """self.in_chans = n_channels
        self.n_classes = n_classes
        self.input_window_samples = n_samples
        self.final_conv_length = final_conv_length
        self.n_filters_time = n_filters_time
        self.n_filters_spat = n_filters_spat
        self.filter_time_length = filter_time_length
        self.pool_time_length = pool_time_length
        self.pool_time_stride = pool_time_stride
        self.n_filters_2 = n_filters_2
        self.filter_length_2 = filter_length_2
        self.n_filters_3 = n_filters_3
        self.filter_length_3 = filter_length_3
        self.n_filters_4 = n_filters_4
        self.filter_length_4 = filter_length_4
        self.first_nonlin = first_nonlin
        self.first_pool_mode = first_pool_mode
        self.first_pool_nonlin = first_pool_nonlin
        self.later_nonlin = later_nonlin
        self.later_pool_mode = later_pool_mode
        self.later_pool_nonlin = later_pool_nonlin
        self.drop_prob = drop_prob
        self.double_time_convs = double_time_convs
        self.split_first_layer = split_first_layer
        self.batch_norm = batch_norm
        self.batch_norm_alpha = batch_norm_alpha
        self.stride_before_pool = stride_before_pool"""

        if stride_before_pool:
            conv_stride = pool_time_stride
            pool_stride = 1
        else:
            conv_stride = 1
            pool_stride = pool_time_stride
        self.add_module("ensuredims", Ensure4d())

        pool_class_dict = dict(max=nn.MaxPool2d, mean=AvgPool2dWithConv)
        first_pool_class = pool_class_dict[first_pool_mode]
        later_pool_class = pool_class_dict[later_pool_mode]
        if split_first_layer:
            self.add_module("dimshuffle", Expression(transpose_time_to_spat))
            self.add_module(
                "conv_time",
                nn.Conv2d(
                    1,
                    n_filters_time,
                    (filter_time_length, 1),
                    stride=1,
                ),
            )
            self.add_module(
                "conv_spat",
                nn.Conv2d(
                    n_filters_time,
                    n_filters_spat,
                    (1, n_channels),
                    stride=(conv_stride, 1),
                    bias=not batch_norm,
                ),
            )
            n_filters_conv = n_filters_spat
        else:
            self.add_module(
                "conv_time",
                nn.Conv2d(
                    n_channels,
                    n_filters_time,
                    (filter_time_length, 1),
                    stride=(conv_stride, 1),
                    bias=not batch_norm,
                ),
            )
            n_filters_conv = n_filters_time
        if batch_norm:
            self.add_module(
                "bnorm",
                nn.BatchNorm2d(
                    n_filters_conv,
                    momentum=batch_norm_alpha,
                    affine=True,
                    eps=1e-5,
                ),
            )
        self.add_module("conv_nonlin", Expression(first_nonlin))
        self.add_module(
            "pool",
            first_pool_class(
                kernel_size=(pool_time_length, 1), stride=(pool_stride, 1)
            ),
        )
        self.add_module("pool_nonlin", Expression(first_pool_nonlin))

        def add_conv_pool_block(n_filters_before, n_filters, filter_length, block_nr):
            suffix = "_{:d}".format(block_nr)
            self.add_module("drop" + suffix, nn.Dropout(p=drop_prob))
            self.add_module(
                "conv" + suffix,
                nn.Conv2d(
                    n_filters_before,
                    n_filters,
                    (filter_length, 1),
                    stride=(conv_stride, 1),
                    bias=not batch_norm,
                ),
            )
            if batch_norm:
                self.add_module(
                    "bnorm" + suffix,
                    nn.BatchNorm2d(
                        n_filters,
                        momentum=batch_norm_alpha,
                        affine=True,
                        eps=1e-5,
                    ),
                )
            self.add_module("nonlin" + suffix, Expression(later_nonlin))

            self.add_module(
                "pool" + suffix,
                later_pool_class(
                    kernel_size=(pool_time_length, 1),
                    stride=(pool_stride, 1),
                ),
            )
            self.add_module("pool_nonlin" + suffix, Expression(later_pool_nonlin))

        add_conv_pool_block(n_filters_conv, n_filters_2, filter_length_2, 2)
        add_conv_pool_block(n_filters_2, n_filters_3, filter_length_3, 3)
        add_conv_pool_block(n_filters_3, n_filters_4, filter_length_4, 4)

        # self.add_module('drop_classifier', nn.Dropout(p=self.drop_prob))
        self.eval()
        if final_conv_length == "auto":
            out = self(
                np_to_th(
                    np.ones(
                        (1, n_channels, n_samples, 1),
                        dtype=np.float32,
                    )
                )
            )
            n_out_time = out.cpu().data.numpy().shape[2]
            final_conv_length = n_out_time
        self.add_module(
            "conv_classifier",
            nn.Conv2d(
                n_filters_4,
                n_classes,
                (final_conv_length, 1),
                bias=True,
            ),
        )
        self.add_module("softmax", nn.LogSoftmax(dim=1))
        self.add_module("squeeze", Expression(squeeze_final_output))

        # Initialization, xavier is same as in our paper...
        # was default from lasagne
        init.xavier_uniform_(self.conv_time.weight, gain=1)
        # maybe no bias in case of no split layer and batch norm
        if split_first_layer or (not self.batch_norm):
            init.constant_(self.conv_time.bias, 0)
        if split_first_layer:
            init.xavier_uniform_(self.conv_spat.weight, gain=1)
            if not batch_norm:
                init.constant_(self.conv_spat.bias, 0)
        if batch_norm:
            init.constant_(self.bnorm.weight, 1)
            init.constant_(self.bnorm.bias, 0)
        param_dict = dict(list(self.named_parameters()))
        for block_nr in range(2, 5):
            conv_weight = param_dict["conv_{:d}.weight".format(block_nr)]
            init.xavier_uniform_(conv_weight, gain=1)
            if not batch_norm:
                conv_bias = param_dict["conv_{:d}.bias".format(block_nr)]
                init.constant_(conv_bias, 0)
            else:
                bnorm_weight = param_dict["bnorm_{:d}.weight".format(block_nr)]
                bnorm_bias = param_dict["bnorm_{:d}.bias".format(block_nr)]
                init.constant_(bnorm_weight, 1)
                init.constant_(bnorm_bias, 0)

        init.xavier_uniform_(self.conv_classifier.weight, gain=1)
        init.constant_(self.conv_classifier.bias, 0)

        # Start in eval mode
        self.eval()



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/10/10
# License: MIT License
"""
Task Decomposition Component Analysis.
"""
from typing import Optional, List

import numpy as np
from scipy.linalg import qr
from scipy.stats import pearsonr
from numpy import ndarray
from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin

from .cca import FilterBankSSVEP
from .dsp import xiang_dsp_kernel, xiang_dsp_feature


def proj_ref(Yf: ndarray):
    Q, R = qr(Yf.T, mode="economic")
    P = Q @ Q.T
    return P


def aug_2(X: ndarray, n_samples: int, padding_len: int, P: ndarray, training: bool = True):
    X = X.reshape((-1, *X.shape[-2:]))
    n_trials, n_channels, n_points = X.shape
    if n_points < padding_len + n_samples:
        raise ValueError("the length of X should be larger than l+n_samples.")
    aug_X = np.zeros((n_trials, (padding_len + 1) * n_channels, n_samples))
    if training:
        for i in range(padding_len + 1):
            aug_X[:, i * n_channels : (i + 1) * n_channels, :] = X[
                ..., i : i + n_samples
            ]
    else:
        for i in range(padding_len + 1):
            aug_X[:, i * n_channels : (i + 1) * n_channels, : n_samples - i] = X[
                ..., i:n_samples
            ]
    aug_Xp = aug_X @ P
    aug_X = np.concatenate([aug_X, aug_Xp], axis=-1)
    return aug_X


def tdca_feature(
    X: ndarray,
    templates: ndarray,
    W: ndarray,
    M: ndarray,
    Ps: List[ndarray],
    padding_len: int,
    n_components: int = 1,
    training=False,
):
    rhos = []
    for Xk, P in zip(templates, Ps):
        a = xiang_dsp_feature(
            W,
            M,
            aug_2(X, P.shape[0], padding_len, P, training=training),
            n_components=n_components,
        )
        b = Xk[:n_components, :]
        a = np.reshape(a, (-1))
        b = np.reshape(b, (-1))
        rhos.append(pearsonr(a, b)[0])
    return rhos


class TDCA(BaseEstimator, TransformerMixin, ClassifierMixin):
    def __init__(self, padding_len: int, n_components: int = 1):
        self.padding_len = padding_len
        self.n_components = n_components

    def fit(self, X: ndarray, y: ndarray, Yf: ndarray):
        X -= np.mean(X, axis=-1, keepdims=True)
        self.classes_ = np.unique(y)
        self.Ps_ = [proj_ref(Yf[i]) for i in range(len(self.classes_))]

        aug_X_list, aug_Y_list = [], []
        for i, label in enumerate(self.classes_):
            aug_X_list.append(
                aug_2(
                    X[y == label],
                    self.Ps_[i].shape[0],
                    self.padding_len,
                    self.Ps_[i],
                    training=True,
                )
            )
            aug_Y_list.append(y[y == label])

        aug_X = np.concatenate(aug_X_list, axis=0)
        aug_Y = np.concatenate(aug_Y_list, axis=0)
        self.W_, _, self.M_, _ = xiang_dsp_kernel(aug_X, aug_Y)

        self.templates_ = np.stack(
            [
                np.mean(
                    xiang_dsp_feature(
                        self.W_,
                        self.M_,
                        aug_X[aug_Y == label],
                        n_components=self.W_.shape[1],
                    ),
                    axis=0,
                )
                for label in self.classes_
            ]
        )
        return self

    def transform(self, X: ndarray):
        n_components = self.n_components
        X -= np.mean(X, axis=-1, keepdims=True)
        X = X.reshape((-1, *X.shape[-2:]))
        rhos = [
            tdca_feature(
                tmp,
                self.templates_,
                self.W_,
                self.M_,
                self.Ps_,
                self.padding_len,
                n_components=n_components,
            )
            for tmp in X
        ]
        rhos = np.stack(rhos)
        return rhos

    def predict(self, X: ndarray):
        feat = self.transform(X)
        labels = self.classes_[np.argmax(feat, axis=-1)]
        return labels


class FBTDCA(FilterBankSSVEP, ClassifierMixin):
    def __init__(
        self,
        filterbank: List[ndarray],
        padding_len: int,
        n_components: int = 1,
        filterweights: Optional[ndarray] = None,
        n_jobs: Optional[int] = None,
    ):
        self.padding_len = padding_len
        self.n_components = n_components
        self.filterweights = filterweights
        self.n_jobs = n_jobs
        super().__init__(
            filterbank,
            TDCA(padding_len, n_components=n_components),
            filterweights=filterweights,
            n_jobs=n_jobs,
        )

    def fit(self, X: ndarray, y: ndarray, Yf: Optional[ndarray] = None):  # type: ignore[override]
        self.classes_ = np.unique(y)
        super().fit(X, y, Yf=Yf)
        return self

    def predict(self, X: ndarray):
        features = self.transform(X)
        if self.filterweights is None:
            features = np.reshape(
                features, (features.shape[0], len(self.filterbank), -1)
            )
            features = np.mean(features, axis=1)
        labels = self.classes_[np.argmax(features, axis=-1)]
        return labels



from .base_tmpl import BaseTmpl
import os
from metabci.brainda.utils import download

class TestDownload(BaseTmpl):

    def test_download(self):
        tmpdir = "/tests"

        url = "https://www.google.com/data/folder"
        dest = os.path.join(tmpdir, "data", "folder")
        self.dbgPrint(dest)
        self.assertEqual(dest, download._url_to_local_path(url, tmpdir))




# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/3/4
# License: MIT License
"""
Munich MI dataset.
Unkown channel names.
"""
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

from mne.io import Raw, read_raw_eeglab
from .base import BaseDataset
from ..utils.download import mne_data_path
from ..utils.channels import upper_ch_names

MUNICH_URL = "https://zenodo.org/record/1217449/files/"


class MunichMI(BaseDataset):
    """Munich Motor Imagery dataset.

    Motor imagery dataset from Grosse-Wentrup et al. 2009 [1]_.

    A trial started with the central display of a white fixation cross. After 3
    s, a white arrow was superimposed on the fixation cross, either pointing to
    the left or the right.
    Subjects were instructed to perform haptic motor imagery of the
    left or the right hand during display of the arrow, as indicated by the
    direction of the arrow. After another 7 s, the arrow was removed,
    indicating the end of the trial and start of the next trial. While subjects
    were explicitly instructed to perform haptic motor imagery with the
    specified hand, i.e., to imagine feeling instead of visualizing how their
    hands moved, the exact choice of which type of imaginary movement, i.e.,
    moving the fingers up and down, gripping an object, etc., was left
    unspecified.
    A total of 150 trials per condition were carried out by each subject,
    with trials presented in pseudorandomized order.

    Ten healthy subjects (S1–S10) participated in the experimental
    evaluation. Of these, two were females, eight were right handed, and their
    average age was 25.6 years with a standard deviation of 2.5 years. Subject
    S3 had already participated twice in a BCI experiment, while all other
    subjects were naive to BCIs. EEG was recorded at M=128 electrodes placed
    according to the extended 10–20 system. Data were recorded at 500 Hz with
    electrode Cz as reference. Four BrainAmp amplifiers were used for this
    purpose, using a temporal analog high-pass filter with a time constant of
    10 s. The data were re-referenced to common average reference
    offline. Electrode impedances were below 10 kΩ for all electrodes and
    subjects at the beginning of each recording session. No trials were
    rejected and no artifact correction was performed. For each subject, the
    locations of the 128 electrodes were measured in three dimensions using a
    Zebris ultrasound tracking system and stored for further offline analysis.


    References
    ----------
    .. [1] Grosse-Wentrup, Moritz, et al. "Beamforming in noninvasive
           brain–computer interfaces." IEEE Transactions on Biomedical
           Engineering 56.4 (2009): 1209-1219.

    """

    _EVENTS = {
        "left_hand": (10, (0, 7)),
        "right_hand": (20, (0, 7)),
    }

    _CHANNELS = [str(i) for i in range(1, 129)]

    def __init__(self):
        super().__init__(
            dataset_code="munichmi",
            subjects=list(range(1, 11)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=500,
            paradigm="imagery",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        url = "{:s}subject{:d}.fdt".format(MUNICH_URL, subject)
        mne_data_path(
            url,
            self.dataset_code,
            path=path,
            proxies=proxies,
            force_update=force_update,
            update_path=update_path,
        )

        url = "{:s}subject{:d}.set".format(MUNICH_URL, subject)
        dests = [
            [
                mne_data_path(
                    url,
                    self.dataset_code,
                    path=path,
                    proxies=proxies,
                    force_update=force_update,
                    update_path=update_path,
                )
            ]
        ]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)

        # DON'T KNOW CHANNEL NAMES!!!
        # montage = make_standard_montage('standard_1005')
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        sess = dict()
        for isess, run_dests in enumerate(dests):
            runs = dict()
            for irun, run_array in enumerate(run_dests):
                raw = read_raw_eeglab(run_array, preload=True)
                raw = upper_ch_names(raw)
                # raw.set_montage(montage)
                raw.annotations.delete(0)
                runs["run_{:d}".format(irun)] = raw
            sess["session_{:d}".format(isess)] = runs
        return sess



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/9/12
# License: MIT License
"""
Conv-CA
Modified from https://github.com/yaoli90/Conv-CA

"""
from collections import OrderedDict

import torch
import torch.nn as nn

from .base import _glorot_weight_zero_bias, compute_same_pad2d, SkorchNet


class _CorrLayer(nn.Module):
    def __init__(self):
        super(_CorrLayer, self).__init__()

    def forward(self, X, T):
        # X: n_batch, 1, 1, n_samples
        # T: n_batch, 1, n_classes, n_samples
        T = torch.swapaxes(T, -1, -2)
        corr_xt = torch.matmul(X, T)  # n_batch, 1, 1, n_classes
        corr_xx = torch.sum(torch.square(X), -1, keepdim=True)
        corr_tt = torch.sum(torch.square(T), -2, keepdim=True)
        corr = corr_xt / (torch.sqrt(corr_xx) * torch.sqrt(corr_tt))
        return corr


@SkorchNet
class ConvCA(nn.Module):
    def __init__(self, n_channels, n_samples, n_classes):
        # super(ConvCA, self).__init__()
        super().__init__()

        n_time_filters_signal = 16
        n_time_filters_template = 40
        time_conv_kernel = 9
        dropout_signal = 0.75
        dropout_template = 0.15

        # input: n_batch, 1, n_channel, n_samples
        self.signal_cnn = nn.Sequential(
            OrderedDict(
                [
                    (
                        "same_padding1",
                        nn.ConstantPad2d(
                            compute_same_pad2d(
                                (n_channels, n_samples),
                                (n_channels, time_conv_kernel),
                                stride=(1, 1),
                            ),
                            0,
                        ),
                    ),
                    (
                        "conv1",
                        nn.Conv2d(
                            1,
                            n_time_filters_signal,
                            (n_channels, time_conv_kernel),
                            stride=(1, 1),
                            padding=0,
                            bias=True,
                        ),
                    ),
                    (
                        "same_padding2",
                        nn.ConstantPad2d(
                            compute_same_pad2d(
                                (n_channels, n_samples), (n_channels, 1), stride=(1, 1)
                            ),
                            0,
                        ),
                    ),
                    (
                        "conv2",
                        nn.Conv2d(
                            n_time_filters_signal,
                            1,
                            (n_channels, 1),
                            stride=(1, 1),
                            padding=0,
                            bias=True,
                        ),
                    ),
                    (
                        "conv3",
                        nn.Conv2d(
                            1, 1, (n_channels, 1), stride=(1, 1), padding=0, bias=True
                        ),
                    ),
                    ("dropout", nn.Dropout(dropout_signal)),
                ]
            )
        )

        # input: n_batch, n_channels, n_classes, n_samples
        self.template_cnn = nn.Sequential(
            OrderedDict(
                [
                    (
                        "same_padding1",
                        nn.ConstantPad2d(
                            compute_same_pad2d(
                                (n_classes, n_samples),
                                (1, time_conv_kernel),
                                stride=(1, 1),
                            ),
                            0,
                        ),
                    ),
                    (
                        "conv1",
                        nn.Conv2d(
                            n_channels,
                            n_time_filters_template,
                            (1, time_conv_kernel),
                            stride=(1, 1),
                            padding=0,
                            bias=True,
                        ),
                    ),
                    (
                        "same_padding2",
                        nn.ConstantPad2d(
                            compute_same_pad2d(
                                (n_classes, n_samples),
                                (1, time_conv_kernel),
                                stride=(1, 1),
                            ),
                            0,
                        ),
                    ),
                    (
                        "conv2",
                        nn.Conv2d(
                            n_time_filters_template,
                            1,
                            (1, time_conv_kernel),
                            stride=(1, 1),
                            padding=0,
                            bias=True,
                        ),
                    ),
                    ("dropout", nn.Dropout(dropout_template)),
                ]
            )
        )

        self.corr_layer = _CorrLayer()
        self.flatten_layer = nn.Flatten()
        self.fc_layer = nn.Linear(n_classes, n_classes)
        self._reset_parameters()

    @torch.no_grad()
    def _reset_parameters(self):
        _glorot_weight_zero_bias(self)

    def forward(self, X, T):
        # X: (n_batch, n_channels, n_samples)
        # T: (n_batch, n_channels, n_classes, n_samples)
        X = X.unsqueeze(1)
        X = self.signal_cnn(X)
        T = self.template_cnn(T)
        corr = self.corr_layer(X, T)
        corr = self.flatten_layer(corr)
        out = self.fc_layer(corr)
        return out
